/*
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see  <http://www.gnu.org/licenses/>
 */

#include "AmaralModel.hxx"

#include "Problem.hxx"
#include "Department.hxx"
#include "Constraint.hxx"
#include "Variable.hxx"

#include <algorithm>

AmaralModel::AmaralModel():
    BaseModel(*static_cast<Problem*>(nullptr), false)
{
    // THIS IS NEVER CALLED
    // This constructor only exists because of an (old) gcc bug with
    // virtual inheritance
    assert(0);
}

AmaralModel::AmaralModel(Problem& problem, bool use_cuts):
    BaseModel(problem, use_cuts)
{
}

void AmaralModel::init_p_variables()
{
    size_type n = _problem.get_num_orig();
    _p_variables.resize( n );
    for ( size_type i = 0; i < n; ++i ) {
        Variable* var = new Variable(department(i),
                                     _variables.size(),
                                     "p", Variable::Cont);
        _p_variables[i] = var;
        _variables.push_back(var);
    }
}

void AmaralModel::init_a_variables()
{
    size_type n = _problem.get_num_orig();
    _a_variables.resize( n );
    for ( size_type i = 0; i < n; ++i ) {
        _a_variables[i].assign( n, NULL );
    }
    for ( size_type i = 0; i < n; ++i ) {
        for ( size_type j = 0; j < n; ++j ) {
            if ( i == j ) continue;
            Variable* var = new Variable(department(i),
                                         department(j),
                                         _variables.size(),
                                         "a", Variable::Binary);
            _a_variables[i][j] = var;
            _variables.push_back(var);
        }
    }
}



void AmaralModel::init_amaral_constraints()
{
    size_type n = _problem.get_num_orig();
    for ( size_type i = 0; i < n; i++ ){
        for ( size_type j = i+1; j < n; j++ ){
            for ( size_type k = 0; k < n; k++ ){
                if ( k == i || k == j ) continue;
                Constraint* constraint = new Constraint( Constraint::LE, 1.);
                constraint->set_coeff( _a_variables[i][j], -1.0 );
                constraint->set_coeff( _a_variables[i][k], 1.0 );
                constraint->set_coeff( _a_variables[j][k], 1.0 );
                constraint->set_coeff( _a_variables[j][i], -1.0 );
                constraint->set_coeff( _a_variables[k][i], 1.0 );
                constraint->set_coeff( _a_variables[k][j], 1.0 );
                _constraints.push_back(constraint);
            }
        }
    }
    for ( size_type i = 0; i < n; i++ ){
        for ( size_type j = i+1; j < n; j++ ){
            for ( size_type k = 0; k < j; k++ ){
                if ( k == i ) continue;
                Constraint* constraint = new Constraint( Constraint::LE, 1.);
                constraint->set_coeff( _a_variables[i][j], -1.0 );
                constraint->set_coeff( _a_variables[i][k], 1.0 );
                constraint->set_coeff( _a_variables[j][k], -1.0 );
                constraint->set_coeff( _a_variables[j][i], 1.0 );
                constraint->set_coeff( _a_variables[k][i], -1.0 );
                constraint->set_coeff( _a_variables[k][j], 1.0 );
                _constraints.push_back(constraint);
            }
        }
    }
    for ( size_type i = 0; i < n; i++ ){
        for ( size_type j = i+1; j < n; j++ ){
            for ( size_type k = j+1; k < n; k++ ){
                Constraint* constraint = new Constraint( Constraint::GE, 1.);
                constraint->set_coeff( _a_variables[i][j], 1.0 );
                constraint->set_coeff( _a_variables[i][k], 1.0 );
                constraint->set_coeff( _a_variables[j][k], 1.0 );
                constraint->set_coeff( _a_variables[j][i], 1.0 );
                constraint->set_coeff( _a_variables[k][i], 1.0 );
                constraint->set_coeff( _a_variables[k][j], 1.0 );
                _constraints.push_back(constraint);
            }
        }
    }
}



std::vector<std::vector<ID> > AmaralModel::compute_solution(const std::vector<double>& x)
{
    size_type n = _problem.get_num_orig();
    std::vector<std::vector<ID>> sol;

    for (size_type j = 0; j < n; j++) {
        for (size_type k = 0; k <= sol.size(); k++) {
            if (k == sol.size()) {
                sol.push_back({j});
                break;
            } else {
                auto var1 = _a_variables[sol[k][0]][j];
                auto var2 = _a_variables[j][sol[k][0]];
                if (var1 && var2 && (x[var1->get_index()] > 0.5 || x[var2->get_index()] > 0.5)) {
                    sol[k].push_back(j);
                    break;
                }
            }
        }
    }

    for (auto& row : sol) {
        std::sort(row.begin(), row.end(),
             [&] (ID i, ID j) {
                 auto var = _a_variables[i][j];
                 return x[var->get_index()] > 0.5;
             });
    }

    return sol;
}
