/*
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see  <http://www.gnu.org/licenses/>
 */

#ifndef __BASEMODEL_HXX__
#define __BASEMODEL_HXX__

#include "Types.hxx"

#include <vector>

class Problem;
class Department;
class Variable;
class Constraint;
class Separator;

// our basic model

class BaseModel
{
protected:
    Problem& _problem;

    std::vector<Variable*> _variables;
    std::vector<Constraint*> _constraints;

    /// distance variables
    std::vector<std::vector<Variable*> > _d_variables;

private:
    /// whether the model separates cuts
    bool _use_cuts;

public:
    BaseModel(Problem& problem, bool use_cuts);

    virtual ~BaseModel();

    /// Return problem of this model.
    const Problem& get_problem() const { return _problem; }

    /// Initialize the model.
    virtual void init() = 0;

    /// Return whether the model uses cuts.
    bool use_cuts() const { return _use_cuts; }

    /// Return an upper bound.
    virtual double get_upper() const;

    /// Return the number of variables in this model.
    size_type get_num_variables() const { return _variables.size(); }

    /// Return a variable.
    const Variable* get_variable(size_type i) const {
        return _variables[i];
    }

    /// Return the number of constraints in this model.
    size_type get_num_constraints() const { return _constraints.size(); }

    /// Return a constraint.
    const Constraint* get_constraint(size_type i) const {
        return _constraints[i];
    }


    /// Called to separate some constraints.
    virtual void separate(const std::vector<double>& x, Separator& separator);

    /**
     * Return solution.
     *
     * @param x the variable values.
     * @return List of departments for each row.
     */
    virtual std::vector<std::vector<ID> >
    compute_solution(const std::vector<double>& x) = 0;

protected:
    /**
     * Initialize distance variables.
     *
     * Creates variables for pairs of departments 0 <= i < j < n,
     * where n is a parameter depending on the model with n >= the
     * number of departments.
     *
     * @param n the number of departments
     */
    void init_d_variables(size_type n);

    /// Return the department with given id.
    virtual const Department* department(ID id) const;

    /// Return the symmetric transport between two departments.
    virtual double get_transport(const Department* dep1, const Department* dep2) const;
};


#endif
