/*
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see  <http://www.gnu.org/licenses/>
 */

#include "BetweennessModel.hxx"

#include "Problem.hxx"
#include "Department.hxx"
#include "Constraint.hxx"
#include "Variable.hxx"
#include "Separator.hxx"
#include "InitSeparator.hxx"

#include <cmath>
#include <algorithm>


BetweennessModel::BetweennessModel(Problem& problem, const rows_vec& rows, bool use_cuts, double upper):
    BaseModel(problem, use_cuts), _rows(rows), _upper(upper)
{
    _dep_left = new Department();
    _dep_left->set_length(0);
    _dep_left->set_id(problem.get_num_orig());

    _dep_right = new Department();
    _dep_right->set_length(0);
    _dep_right->set_id(problem.get_num_orig()+1);
}

BetweennessModel::~BetweennessModel()
{
    delete _dep_left;
    delete _dep_right;
}

const Department* BetweennessModel::department(ID i) const
{
    size_type n = _problem.get_num_orig();

    assert(i < n + 2);

    if (i < n)
        return _problem.get_department(i);
    else if (i == n)
        return _dep_left;
    else
        return _dep_right;
}

double BetweennessModel::get_transport(const Department* dep1, const Department* dep2) const
{
    size_type n = _problem.get_num_orig();
    if (dep1->get_id() < n && dep2->get_id() < n)
        return BaseModel::get_transport(dep1, dep2);
    else
        return 0;
}

void BetweennessModel::init_x_variables()
{
    size_type n = _problem.get_num_orig() + 2;

    _x_variables.resize( n );
    for ( size_type i = 0; i < n; ++i ) {
        _x_variables[i].resize( n );
        for ( size_type j = 0; j < n; ++j ) {
            _x_variables[i][j].assign( n, NULL );
        }
    }
    for ( size_type m = 0; m < _rows.size(); m++ ){
        for ( size_type i = 0; i < _rows[m].size()+2; i++){
            for ( size_type j = 0; j < _rows[m].size()+2; j++){
                if ( i == j ) continue;
                for ( size_type k = i+1; k < _rows[m].size()+2; k++){
                    if (j == k) continue;
                    size_type idx_i, idx_j, idx_k;

                    if      ( i <  _rows[m].size() ) idx_i = _rows[m][i];
                    else if ( i == _rows[m].size() ) idx_i = n-2;
                    else                             idx_i = n-1;

                    if      ( j <  _rows[m].size() ) idx_j = _rows[m][j];
                    else if ( j == _rows[m].size() ) idx_j = n-2;
                    else                             idx_j = n-1;

                    if      ( k <  _rows[m].size() ) idx_k = _rows[m][k];
                    else if ( k == _rows[m].size() ) idx_k = n-2;
                    else                             idx_k = n-1;

                    Variable* var = new Variable(department(idx_i),
                                                 department(idx_j),
                                                 department(idx_k),
                                                 _variables.size(),
                                                 "x", Variable::Binary);
                    _x_variables[idx_i][idx_j][idx_k] = var;
                    _x_variables[idx_k][idx_j][idx_i] = var;
                    _variables.push_back(var);
                }
            }
        }
    }
}


void BetweennessModel::init_betweenness_constraints()
{
    size_type n = _problem.get_num_orig() + 2;

    for ( size_type m = 0; m < _rows.size(); m++ ){
        for ( size_type i = 0; i < _rows[m].size()+2; i++){
            for ( size_type j = i+1; j < _rows[m].size()+2; j++){
                for ( size_type k = j+1; k < _rows[m].size()+2 ; k++){
                    size_type idx_i, idx_j, idx_k;
                    if ( i < _rows[m].size() ) idx_i = _rows[m][i];
                    if ( j < _rows[m].size() ) idx_j = _rows[m][j];
                    if ( k < _rows[m].size() ) idx_k = _rows[m][k];
                    if ( i == _rows[m].size() ) idx_i = n-2;
                    if ( i == _rows[m].size()+1 ) idx_i = n-1;
                    if ( j == _rows[m].size() ) idx_j = n-2;
                    if ( j == _rows[m].size()+1 ) idx_j = n-1;
                    if ( k == _rows[m].size() ) idx_k = n-2;
                    if ( k == _rows[m].size()+1 ) idx_k = n-1;
                    Constraint* constraint = new Constraint( Constraint::EQ, 1.);
                    constraint->set_coeff( _x_variables[idx_i][idx_j][idx_k], 1.0 );
                    constraint->set_coeff( _x_variables[idx_i][idx_k][idx_j], 1.0 );
                    constraint->set_coeff( _x_variables[idx_j][idx_i][idx_k], 1.0 );
                    _constraints.push_back(constraint);
                }
            }
        }
    }

    for ( size_type i = 0; i < n-2; i++ ){
        Constraint* constraint = new Constraint( Constraint::EQ, 1.);
        constraint->set_coeff( _x_variables[n-2][i][n-1], 1.0 );
        _constraints.push_back(constraint);
    }
    for ( size_type i = 0; i < n-2; i++ ){
        Constraint* constraint = new Constraint( Constraint::EQ, 0.);
        constraint->set_coeff( _x_variables[n-2][n-1][i], 1.0 );
        _constraints.push_back(constraint);
    }
    for ( size_type i = 0; i < n-2; i++ ){
        Constraint* constraint = new Constraint( Constraint::EQ, 0.);
        constraint->set_coeff( _x_variables[i][n-2][n-1], 1.0 );
        _constraints.push_back(constraint);
    }

    for ( size_type m = 0; m < _rows.size(); m++ ) {
        for ( size_type i = 0; i < _rows[m].size(); i++){
            for ( size_type j = i+1; j < _rows[m].size(); j++){
                {
                    Constraint* constraint = new Constraint( Constraint::EQ, 0.);
                    constraint->set_coeff( _x_variables[_rows[m][i]][n-2][_rows[m][j]], 1.0 );
                    _constraints.push_back(constraint);
                }
                {
                    Constraint* constraint = new Constraint( Constraint::EQ, 0.);
                    constraint->set_coeff( _x_variables[_rows[m][i]][n-1][_rows[m][j]], 1.0 );
                    _constraints.push_back(constraint);
                }
                {
                    Constraint* constraint = new Constraint( Constraint::EQ, 0.);
                    constraint->set_coeff( _x_variables[_rows[m][i]][_rows[m][j]][n-2], 1.0 );
                    constraint->set_coeff( _x_variables[_rows[m][j]][_rows[m][i]][n-1], -1.0 );
                    _constraints.push_back(constraint);
                }
            }
        }
    }

    if ( use_cuts() == false ) {
        InitSeparator isep(_constraints);
        separate_betweenness_constraints(isep);
    }
}


void BetweennessModel::separate_betweenness_constraints(Separator& separator)
{
    size_type n = _problem.get_num_orig() + 2;
    for ( size_type m = 0; m < _rows.size(); m++ ){
        for ( size_type i = 0; i < _rows[m].size()+2; i++){
            for ( size_type j = i+1; j < _rows[m].size()+2; j++){
                for ( size_type k = j+1; k < _rows[m].size()+2; k++){
                    for ( size_type d = 0; d < _rows[m].size(); d++){
                        if ( d == i || d == j || d == k ) continue;
                        size_type idx_i, idx_j, idx_k, idx_d;
                        if ( i < _rows[m].size() ) idx_i = _rows[m][i];
                        if ( j < _rows[m].size() ) idx_j = _rows[m][j];
                        if ( k < _rows[m].size() ) idx_k = _rows[m][k];
                        idx_d = _rows[m][d];
                        if ( i == _rows[m].size() ) idx_i = n-2;
                        if ( i == _rows[m].size()+1 ) idx_i = n-1;
                        if ( j == _rows[m].size() ) idx_j = n-2;
                        if ( j == _rows[m].size()+1 ) idx_j = n-1;
                        if ( k == _rows[m].size() ) idx_k = n-2;
                        if ( k == _rows[m].size()+1 ) idx_k = n-1;
                        Separator::coeff_vec vars;
                        vars.push_back( std::make_pair( _x_variables[idx_i][idx_d][idx_j], 1.0) );
                        vars.push_back( std::make_pair( _x_variables[idx_i][idx_d][idx_k], 1.0) );
                        vars.push_back( std::make_pair( _x_variables[idx_j][idx_d][idx_k], 1.0) );
                        separator.add_if_violated( vars, Constraint::LE, 2.0);
                        vars.clear();
                        vars.push_back( std::make_pair( _x_variables[idx_i][idx_d][idx_j], -1.0) );
                        vars.push_back( std::make_pair( _x_variables[idx_i][idx_d][idx_k], 1.0) );
                        vars.push_back( std::make_pair( _x_variables[idx_j][idx_d][idx_k], 1.0) );
                        separator.add_if_violated( vars, Constraint::GE, 0.0);
                        vars.clear();
                        vars.push_back( std::make_pair( _x_variables[idx_i][idx_d][idx_j], 1.0) );
                        vars.push_back( std::make_pair( _x_variables[idx_i][idx_d][idx_k], -1.0) );
                        vars.push_back( std::make_pair( _x_variables[idx_j][idx_d][idx_k], 1.0) );
                        separator.add_if_violated( vars, Constraint::GE, 0.0);
                        vars.clear();
                        vars.push_back( std::make_pair( _x_variables[idx_i][idx_d][idx_j], 1.0) );
                        vars.push_back( std::make_pair( _x_variables[idx_i][idx_d][idx_k], 1.0) );
                        vars.push_back( std::make_pair( _x_variables[idx_j][idx_d][idx_k], -1.0) );
                        separator.add_if_violated( vars, Constraint::GE, 0.0);
                        vars.clear();
                    }
                }
            }
        }
    }
}


void BetweennessModel::separate_triangle_constraints(Separator& separator,
                                                     size_type min_dist)
{
    size_type n = _problem.get_num_orig() + 2;
    for ( size_type i = 0; i < n-1; i++){
        for ( size_type j = i+1; j < n-1; j++){
            for ( size_type k = 0; k < n-1; k++){
                if ( k == j || k == i ) continue;
                if (min_dist >= 2) {
                    int rowi = department(i)->get_row();
                    int rowj = department(j)->get_row();
                    int rowk = department(k)->get_row();
                    if ( std::abs(rowi - rowj) >= (int)min_dist ) continue;
                    if ( std::abs(rowi - rowk) >= (int)min_dist ) continue;
                    if ( std::abs(rowj - rowk) >= (int)min_dist ) continue;
                }
                Separator::coeff_vec vars;
                vars.push_back( std::make_pair( _d_variables[i][j], -1.0) );
                vars.push_back( std::make_pair( _d_variables[i][k], 1.0) );
                vars.push_back( std::make_pair( _d_variables[k][j], 1.0) );
                separator.add_if_violated( vars, Constraint::GE, 0.0);
                vars.clear();
            }
        }
    }
}


double BetweennessModel::get_upper() const
{
    return _upper;
}


std::vector<std::vector<ID> >
BetweennessModel::compute_solution(const std::vector<double>& x)
{
    auto n = _problem.get_num_orig();
    auto sol = _rows;

    for (auto& row : sol) {
        std::sort(row.begin(), row.end(),
                  [&] (ID i, ID j) {
                      auto var = _x_variables[n][i][j];
                      return x[var->get_index()] > 0.5;
                  });
    }

    return sol;
}
