/*
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see  <http://www.gnu.org/licenses/>
 */

#ifndef __BETWEENNESSMODEL_HXX__
#define __BETWEENNESSMODEL_HXX__

#include "BaseModel.hxx"

class Department;

class BetweennessModel: public BaseModel
{
protected:
    /// betweenness variables
    std::vector<std::vector<std::vector<Variable*> > > _x_variables;

    typedef std::vector<std::vector<size_type> > rows_vec;
    /// current fixed row assignment
    rows_vec _rows;

    /// best known upper bound
    double _upper;

    /// special department n+1
    Department* _dep_left;
    /// special department n+2
    Department* _dep_right;

public:
    BetweennessModel(Problem& problem, const rows_vec& rows, bool use_cuts, double upper);

    ~BetweennessModel();

    double get_upper() const;

    std::vector<std::vector<ID> > compute_solution(const std::vector<double>& x);

protected:
    /**
     * Return the department with id i.
     *
     * If i == n or i == n+1, one of the special departments is returned.
     */
    const Department* department(ID i) const;

    /// Return the transport between two (possibly extended) departments.
    double get_transport(const Department* dep1, const Department* dep2) const;

    void init_x_variables();
    void init_betweenness_constraints();
    void separate_betweenness_constraints(Separator& separator);

    /**
     * Separate triangle inequalities.
     *
     * The inequalities are only separated, if the pairwise row
     * distance between the departments is less than `min_dist`.
     *
     * @param separator the separator callback
     * @param min_dist minimal pairwise row distance
     */
    void separate_triangle_constraints(Separator& separator, size_type min_dist = 1);
};

#endif
