/*
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see  <http://www.gnu.org/licenses/>
 */

#include "BnBSolver.hxx"

#include "Problem.hxx"
#include "Department.hxx"
#include "BetweennessModel.hxx"
#include "MIPSolver.hxx"

#include <cmath>
#include <cstdlib>
#include <limits>
#include <iostream>
#include <memory>

BnBSolver::BnBSolver(Problem& problem, size_type nrows, bool use_cuts, MIPSolver& solver):
    _problem(problem), _nrows(nrows), _use_cuts(use_cuts), _solver(solver)
{
    if (nrows <= 0) {
        std::cerr << "Number of rows must be >= 1" << std::endl;
        std::exit(EXIT_FAILURE);
    }
}

double BnBSolver::solve()
{
    size_type n = _problem.get_num_orig() + 2;

    std::vector<size_type> r(n-2, 0);
    rows_vec rows;
    std::unique_ptr<BetweennessModel> submodel = nullptr;

    double upper =  std::numeric_limits<double>::max();
    std::vector<std::vector<ID>> upper_rows;

    // test all row assignments
    bool is_finished = false;
    size_type i = 0;

    while (!is_finished) {

        if ( i % 1000000 == 0 ) std::cout << "i=" << i << std::endl;
        i++;

        // set rows
        rows.clear();
        rows.resize(_nrows);

        for ( size_type j = 0; j < r.size(); j++ ){
            rows[r[j]].push_back(j);
        }

        // check length
        std::vector<double> lens(_nrows, 0.0);
        double l0max = 0.0;

        for ( size_type j = 0; j < r.size(); j++ ) {
            lens[r[j]] += _problem.get_department(j)->get_length();
            if (r[j] == 0) {
                l0max = std::max(l0max, _problem.get_department(j)->get_length());
            }
        }

        bool valid = true;

        for ( size_type j = 1; j < lens.size() && valid; j++ ) {
            if (lens[j] - lens[j-1] > 0.001) valid = false;
        }

        if (check_valid(lens, l0max) == false) valid = false;

        for ( size_type i1 = 0; i1 < _nrows && valid; i1++ ) {
            auto min1 = n;
            for ( auto j : rows[i1] ) min1 = std::min(min1, j);
            for ( size_type i2 = i1+1; i2 < _nrows && valid; i2++) {
                if (fabs(lens[i1] - lens[i2]) > 0.001) continue;
                for ( auto j : rows[i2] ) {
                    if (j < min1) {
                        valid = false;
                        break;
                    }
                }
            }
        }

        if ( valid == true ){
            // build lp und solve
            submodel = create_submodel(rows, upper);

            double val = _solver.solve(*submodel);

            // check if improvement
            if ( val < upper ){
                upper = val;
                std::cout << "Upper bound: " << upper << std::endl;
                upper_rows = submodel->compute_solution(_solver.get_solution());
            }
        }

        // go to next assignment
        for ( size_type j = 0; j < r.size(); j++){
            r[j] += 1;
            // check if there is an overflow
            if (r[j] < _nrows) break;

            // have overflow, stop if we reached the last machine
            r[j] = 0;
            is_finished = j + 1 == r.size();
        }
    }

    _best_rows = std::move(upper_rows);

    return upper;
}
