/*
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see  <http://www.gnu.org/licenses/>
 */

#include "CmdArgs.hxx"

#include <sstream>
#include <ostream>
#include <cstdlib>

namespace lyutil {

CmdArgs::Option::Option(char sopt,
                        const std::string& arg_description,
                        const std::string& description,
                        const std::string& default_value,
                        ArgType type):
    has_short(true), has_long(false),
    sopt(sopt),
    def_value(default_value),
    arg_descr(arg_description),
    type(type),
    description(description),
    is_given(false),
    has_arg(false)
{
}

CmdArgs::Option::Option(const std::string& opt,
                        const std::string& arg_description,
                        const std::string& description,
                        const std::string& default_value,
                        ArgType type):
    has_short(false), has_long(true),
    opt(opt),
    def_value(default_value),
    arg_descr(arg_description),
    type(type),
    description(description),
    is_given(false),
    has_arg(false)
{
}

CmdArgs::Option::Option(char sopt, const std::string& opt,
                        const std::string& arg_description,
                        const std::string& description,
                        const std::string& default_value,
                        ArgType type):
    has_short(true), has_long(true),
    sopt(sopt), opt(opt),
    def_value(default_value),
    arg_descr(arg_description),
    type(type),
    description(description),
    is_given(false),
    has_arg(false)
{
}

CmdArgs::Option::Option(const std::string& title):
    has_short(false), has_long(false),
    description(title)
{
}


CmdArgs::Option::~Option()
{
}


CmdArgs::CmdArgs()
{
}

CmdArgs::Option&
CmdArgs::add(char sopt, const std::string& description) {
    _options.push_back(Option(sopt, "", description, "", Option::NOARG));
    return _options.back();
}

CmdArgs::Option&
CmdArgs::add(const std::string& opt, const std::string& description) {
    _options.push_back(Option(opt, "", description, "", Option::NOARG));
    return _options.back();
}

CmdArgs::Option&
CmdArgs::add(char sopt, const std::string& opt,
             const std::string& description) {
    _options.push_back(Option(sopt, opt, "", description, "",
                              Option::NOARG));
    return _options.back();
}

CmdArgs::Option&
CmdArgs::add_with_optional(const std::string& opt,
                           const std::string& arg_description,
                           const std::string& description,
                           const std::string& default_value) {
    _options.push_back(Option(opt, arg_description, description,
                              default_value, Option::OPTIONAL));
    return _options.back();
}

CmdArgs::Option&
CmdArgs::add_with_optional(char sopt,
                           const std::string& opt,
                           const std::string& arg_description,
                           const std::string& description,
                           const std::string& default_value) {
    _options.push_back(Option(sopt, opt, arg_description, description,
                              default_value, Option::OPTIONAL));
    return _options.back();
}

CmdArgs::Option&
CmdArgs::add_with_required(char sopt,
                           const std::string& arg_description,
                           const std::string& description,
                           const std::string& default_value) {
    _options.push_back(Option(sopt, arg_description, description,
                              default_value, Option::REQUIRED));
    return _options.back();
}

CmdArgs::Option&
CmdArgs::add_with_required(const std::string& opt,
                           const std::string& arg_description,
                           const std::string& description,
                           const std::string& default_value) {
    _options.push_back(Option(opt, arg_description, description,
                              default_value, Option::REQUIRED));
    return _options.back();
}

CmdArgs::Option&
CmdArgs::add_with_required(char sopt, const std::string& opt,
                           const std::string& arg_description,
                           const std::string& description,
                           const std::string& default_value) {
    _options.push_back(Option(sopt, opt, arg_description, description,
                              default_value, Option::REQUIRED));
    return _options.back();
}


void CmdArgs::add_section(const std::string& name) {
    _options.push_back(Option(name));
}


void CmdArgs::parse(int argc, const char** argv) {

    for (int arg_idx = 1; arg_idx < argc; ++arg_idx) {
        std::string arg = argv[arg_idx];

        if (arg.length() == 0) continue;

        if (arg[0] == '-') {
            LYUTIL_TEST( arg.length() > 1,
                         CmdArgError,
                         "CmdArgs::parse",
                         "Options expected after '-'" );

            if (arg[1] == '-') {
                if (arg.length() == 2) {
                    complete_rest(arg_idx+1, argc, argv);
                    break;
                }
                else {
                    arg_idx += parse_long(arg.substr(2), arg_idx,
                                          argc, argv);
                }
            }
            else {
                arg_idx += parse_shorts(arg.substr(1), arg_idx,
                                        argc, argv);
            }
        }
        else
            _rest.push_back(arg);
    }
}


void CmdArgs::complete_rest(int arg_idx, int argc, const char** argv) {
    while (arg_idx < argc) {
        _rest.push_back(argv[arg_idx]);
        arg_idx += 1;
    }
}


int CmdArgs::parse_long(std::string opt, int arg_idx,
                        int argc, const char** argv) {
    int arg_incr = 0;
    bool has_arg = false;

    // first check for --opt=arg style
    std::string::size_type eq_idx = opt.find('=');
    std::string arg;
    if (eq_idx != std::string::npos) {
        arg = opt.substr(eq_idx + 1);
        opt.erase(eq_idx);
        has_arg = true;
    }

    Option* opt_template = find_option(opt);
    LYUTIL_TEST( opt_template != NULL,
                 CmdArgError,
                 "CmdArgs::parse_long",
                 "Option --" << opt << " unknown." );

    LYUTIL_TEST( opt_template->type != Option::NOARG || !has_arg,
                 CmdArgError,
                 "CmdArgs::parse_long",
                 "Option --" << opt << " does not take an argument." );

    if (opt_template->type == Option::REQUIRED && !has_arg) {
        LYUTIL_TEST( arg_idx + 1 < argc,
                     CmdArgError,
                     "CmdArgs::parse_long",
                     "Option --" << opt << " requires an argument." );

        arg = argv[arg_idx+1];
        arg_incr = 1;
        has_arg = true;
    }

    opt_template->has_arg = has_arg;
    opt_template->arg = arg;
    opt_template->is_given = true;

    return arg_incr;
}


int CmdArgs::parse_shorts(const std::string& opt, int arg_idx,
                          int argc, const char** argv) {
    int arg_incr = 0;
    bool has_arg = false;
    std::string arg;

    for (std::string::size_type i = 0; i < opt.size(); ++i) {
        char sopt = opt[i];

        Option* opt_template = find_option(sopt);
        LYUTIL_TEST( opt_template != NULL,
                     CmdArgError,
                     "CmdArgs::parse_shorts",
                     "Option -" << sopt << " unknown." );

        if (opt_template->type == Option::REQUIRED) {
            LYUTIL_TEST( i >= opt.size() - 1 &&
                         arg_idx + 1 < argc,
                         CmdArgError,
                         "CmdArgs::parse_shorts",
                         "Option -" << sopt << " requires an argument." );

            arg_incr = 1;
            has_arg = true;
            arg = argv[arg_idx + 1];
        }

        opt_template->has_arg = has_arg;
        opt_template->arg = arg;
        opt_template->is_given = true;
    }

    return arg_incr;
}


CmdArgs::Option* CmdArgs::find_option(char sopt) {
    for (std::vector<Option>::iterator it = _options.begin();
         it != _options.end(); ++it) {
        if ((*it).has_short && (*it).sopt == sopt)
            return &(*it);
    }
    return NULL;
}


CmdArgs::Option* CmdArgs::find_option(const std::string& opt) {
    for (std::vector<Option>::iterator it = _options.begin();
         it != _options.end(); ++it) {
        if ((*it).has_long && (*it).opt == opt)
            return &(*it);
    }
    return NULL;
}


bool CmdArgs::has(char sopt) const {
    const Option* opt_template = find_option(sopt);

    LYUTIL_TEST( opt_template != NULL,
                 CmdArgError,
                 "CmdArgs::has(char)",
                 "CmdArgError: Unknown option -" << sopt );

    return opt_template->is_given;
}

bool CmdArgs::has(const std::string& opt) const {
    const Option* opt_template = find_option(opt);

    LYUTIL_TEST( opt_template != NULL,
                 CmdArgError,
                 "CmdArgs::has(string)",
                 "CmdArgError: Unknown option --" << opt );

    return opt_template->is_given;
}

bool CmdArgs::has_arg(char sopt) const {
    const Option* opt_template = find_option(sopt);

    LYUTIL_TEST( opt_template != NULL,
                 CmdArgError,
                 "CmdArgs::has_arg(char)",
                 "CmdArgError: Unknown option -" << sopt );

    return opt_template->has_arg;
}

bool CmdArgs::has_arg(const std::string& opt) const {
    const Option* opt_template = find_option(opt);

    LYUTIL_TEST( opt_template != NULL,
                 CmdArgError,
                 "CmdArgs::has_arg(string)",
                 "CmdArgError: Unknown option --" << opt );

    return opt_template->has_arg;
}

const std::string& CmdArgs::get(char sopt) const {
    const Option* opt_template = find_option(sopt);
    static const std::string empty_string;

    LYUTIL_TEST( opt_template != NULL,
                 CmdArgError,
                 "CmdArgs::get(char)",
                 "CmdArgError: Unknown option -" << sopt );

    if (opt_template->has_arg)
        return opt_template->arg;
    else
        return opt_template->def_value;
}

const std::string& CmdArgs::get(const std::string& opt) const {
    const Option* opt_template = find_option(opt);
    static const std::string empty_string;

    LYUTIL_TEST( opt_template != NULL,
                 CmdArgError,
                 "CmdArgs::get(string)",
                 "CmdArgError: Unknown option --" << opt );

    if (opt_template->has_arg)
        return opt_template->arg;
    else
        return opt_template->def_value;
}

int CmdArgs::get_int(char sopt) const {
    return atoi(get(sopt).c_str());
}

int CmdArgs::get_int(const std::string& opt) const {
    return atoi(get(opt).c_str());
}

double CmdArgs::get_dbl(char sopt) const {
    return atoi(get(sopt).c_str());
}

double CmdArgs::get_dbl(const std::string& opt) const {
    return atoi(get(opt).c_str());
}

std::ostream& operator<<(std::ostream& o, const CmdArgs& args) {
    for (std::vector<CmdArgs::Option>::const_iterator it =
             args._options.begin();
         it != args._options.end(); ++it) {
        o << *it << std::endl;
    }
    return o;
}

std::ostream& operator<<(std::ostream& o, const CmdArgs::Option& opt) {
    if (!opt.has_short && !opt.has_long) {
        o << std::endl
          << " " << opt.description << std::endl;
        return o;
    }

    std::string::size_type i = 0;
    o << "  "; i += 2;
    if (opt.has_short)
        o << '-' << opt.sopt;
    else
        o << "  ";
    i += 2;

    if (opt.has_short && opt.has_long)
        o << ", ";
    else
        o << "  ";
    i += 2;

    if (opt.has_long) {
        o << "--" << opt.opt;
        i += 2 + opt.opt.size();
        if (opt.type == CmdArgs::Option::REQUIRED) {
            o << "=" << opt.arg_descr;
            i += 1 + opt.arg_descr.size();
        }
        else if (opt.type == CmdArgs::Option::OPTIONAL) {
            o << "[=" << opt.arg_descr << "]";
            i += 3 + opt.arg_descr.size();
        }
    }

    if (i < 27) i = 30 - i; else i = 3;

    std::string descr = opt.description;
    if ( !opt.def_value.empty() ) {
        descr += "\n  " + std::string(" (default: ") + opt.def_value + ")";
    }

    std::istringstream s(descr);
    while (!s.eof()) {
        std::string line;
        getline(s, line);

        if (i == 30) o << std::endl;
        o << std::string(i, ' ');
        o << line;
        i = 30;
    }

    return o;
}
}
