/*
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see  <http://www.gnu.org/licenses/>
 */

#ifndef __LYUTIL_CMDARGS_HXX__
#define __LYUTIL_CMDARGS_HXX__

#include <string>
#include <vector>
#include <iosfwd>

#include "Error.hxx"

namespace lyutil {

LYUTIL_DEF_ERROR(CmdArgError)

/**
 * \brief Simple class for parsing command-line arguments.
 */
class CmdArgs {

    class Option;

    friend std::ostream& operator<<(std::ostream& o, const CmdArgs& args);
    friend std::ostream& operator<<(std::ostream& o,
                                    const Option& args);

    /**
     * \brief One commandline argument.
     *
     * When the an commandline option is added an object
     * of this class is returned and its methods may be
     * used to specify the properties of this option.
     */
    class Option {

        friend class lyutil::CmdArgs;
        friend std::ostream& operator<<(std::ostream& o, const Option& opt);

    public:
        enum ArgType { NOARG, REQUIRED, OPTIONAL };

        Option(char sopt,
               const std::string& description,
               const std::string& default_value,
               const std::string& arg_description,
               ArgType type);
        Option(const std::string& opt,
               const std::string& description,
               const std::string& default_value,
               const std::string& arg_description,
               ArgType type);
        Option(char sopt, const std::string& opt,
               const std::string& description,
               const std::string& default_value,
               const std::string& arg_description,
               ArgType type);

        Option(const std::string& name);

        ~Option();

        /**
         * \brief This option requires an additional argument!
         *
         * \param arg_description A short name of this option's
         *                        argument (e.g. FILE)
         *
         * \return this
         */
        Option& required( const std::string& arg_description="" ) {
            type = REQUIRED;
            arg_descr = arg_description;
            return *this;
        }

        /**
         * \brief This option has an optional argument!
         *
         * \param arg_description A short name of this option's
         *                        argument (e.g. FILE)
         *
         * \return this
         */
        Option& optional( const std::string& arg_description="" ) {
            type = OPTIONAL;
            arg_descr = arg_description;
            return *this;
        }

        /**
         * \brief Set the default value of this argument.
         *
         * If this option is not specified on the commandline,
         * the 'get'-methods of CmdArgs return this
         * default value.
         *
         * \param def The default value.
         *
         * \return this
         */
        Option& default_value( const std::string& def ) {
            def_value = def;
            return *this;
        }

    private:
        bool has_short;
        bool has_long;

        char sopt;
        std::string opt;

        std::string def_value;
        std::string arg_descr;

        ArgType type;

        std::string description;

        bool is_given;
        bool has_arg;
        std::string arg;
    };

public:

    CmdArgs();

    /**
     * Add a short option without argument.
     *
     * \param sopt Short option character.
     * \param description Help text for this option.
     *
     * \return The new option.
     */
    Option& add(char sopt,
                const std::string& description);

    /**
     * Add a long option without argument.
     *
     * \param opt Long option.
     * \param description Help text for this option.
     *
     * \return The new option.
     */
    Option& add(const std::string& opt,
                const std::string& description);

    /**
     * Add a short and long option without argument.
     *
     * \param sopt Short option character.
     * \param opt Long option.
     * \param description Help text for this option.
     *
     * \return The new option.
     */
    Option& add(char sopt, const std::string& opt,
                const std::string& description);

    /**
     * Add a long option with optional argument.
     *
     * \param opt Long option.
     * \param arg_description Short (usually upcase) text for the argument.
     * \param description Help text for this option.
     * \param default_value Value of this option if no argument given.
     *
     * \return The new option.
     */
    Option& add_with_optional(const std::string& opt,
                              const std::string& arg_description,
                              const std::string& description,
                              const std::string& default_value);

    /**
     * Add a short and long option with optional argument. Only the
     * long option takes an optional argument in the form
     * --opt=arg.
     *
     * \param sopt Short option character.
     * \param opt Long option.
     * \param arg_description Short (usually upcase) text for the argument.
     * \param description Help text for this option.
     * \param default_value Value of this option if no argument given.
     *
     * \return The new option.
     */
    Option& add_with_optional(char sopt, const std::string& opt,
                              const std::string& arg_description,
                              const std::string& description,
                              const std::string& default_value);


    /**
     * Add a short option with required argument.
     *
     * \param opt Long option.
     * \param arg_description Short (usually upcase) text for the argument.
     * \param description Help text for this option.
     * \param default_value Value of this option the option is not given.
     *
     * \return The new option.
     */
    Option& add_with_required(char sopt,
                              const std::string& arg_description,
                              const std::string& description,
                              const std::string& default_value="");

    /**
     * Add a long option with required argument.
     *
     * \param opt Long option.
     * \param arg_description Short (usually upcase) text for the argument.
     * \param description Help text for this option.
     * \param default_value Value of this option the option is not given.
     *
     * \return The new option.
     */
    Option& add_with_required(const std::string& opt,
                              const std::string& arg_description,
                              const std::string& description,
                              const std::string& default_value="");

    /**
     * Add a short and long option with required argument.
     *
     * \param sopt Short option character.
     * \param opt Long option.
     * \param arg_description Short (usually upcase) text for the argument.
     * \param description Help text for this option.
     * \param default_value Value of this option the option is not given.
     *
     * \return The new option.
     */
    Option& add_with_required(char sopt, const std::string& opt,
                              const std::string& arg_description,
                              const std::string& description,
                              const std::string& default_value="");

    /**
     * Add a section.
     *
     * This is shown as a headline in the usage-output.
     *
     * \param title Name of the section.
     */
    void add_section(const std::string& title);

    /// Parse the command line.
    void parse(int argc, const char** argv);

    /// Returns the unparsed arguments.
    const std::vector<std::string>& rest_arguments() const;

    /// Returns true if the (short) option has been given.
    bool has(char sopt) const;
    /// Returns true if the (long) option has been given.
    bool has(const std::string& sopt) const;

    /**
     * Returns true if the (short) option has been given and has an
     * argument.
     */
    bool has_arg(char sopt) const;
    /**
     * Returns true if the (long) option has been given and has an
     * argument.
     */
    bool has_arg(const std::string& sopt) const;

    /// Returns the argument of a short option.
    const std::string& get(char sopt) const;
    /// Returns the argument of a long option.
    const std::string& get(const std::string& opt) const;

    /// Returns the argument of a short option converted to int.
    int get_int(char sopt) const;
    /// Returns the argument of a long option converted to int.
    int get_int(const std::string& opt) const;

    /// Returns the argument of a short option converted to double.
    double get_dbl(char sopt) const;
    /// Returns the argument of a long option converted to double.
    double get_dbl(const std::string& opt) const;

protected:
    void complete_rest(int arg_idx, int argc, const char** argv);

    int parse_long(std::string opt, int arg_idx,
                   int argc, const char** argv);

    int parse_shorts(const std::string& opt, int arg_idx,
                     int argc, const char** argv);

    Option* find_option(char sopt);

    Option* find_option(const std::string& opt);

    const Option* find_option(char sopt) const {
        return const_cast<CmdArgs*>(this)->find_option(sopt);
    }

    const Option* find_option(const std::string& opt) const {
        return const_cast<CmdArgs*>(this)->find_option(opt);
    }


private:
    std::vector<Option> _options;
    std::vector<std::string> _rest;
};

/// Prints the help-text of the command arguments.
std::ostream& operator<<(std::ostream& o, const CmdArgs& args);

/// Prints the help-text of the given option.
std::ostream& operator<<(std::ostream& o, const CmdArgs::Option& args);
}

#endif
