/*
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see  <http://www.gnu.org/licenses/>
 */

#include "Constraint.hxx"

#include "Variable.hxx"

#include <cmath>
#include <limits>
#include <ostream>
#include <cassert>

static bool less_by_index( const Variable* v1, const Variable* v2 )
{
    return v1->get_index() < v2->get_index();
}


Constraint::Constraint( Sense sense, double rhs, const std::string& name ):
    super_type( less_by_index ),
    _sense(sense),
    _rhs(rhs),
    _name(name)
{}


std::vector<int> Constraint::get_indices() const
{
    std::vector<int> indices;
    indices.reserve( size() );

    for ( const_iterator it = begin(); it != end(); ++it )
        indices.push_back( static_cast<int>(it->first->get_index()) );

    return indices;
}


std::vector<double> Constraint::get_coeffs() const
{
    std::vector<double> coeffs;
    coeffs.reserve( size() );

    for ( const_iterator it = begin(); it != end(); ++it )
        coeffs.push_back( it->second );

    return coeffs;
}


void Constraint::add_coeff( const Variable* var, double coeff )
{
    assert( var != NULL );
    iterator it = find(var);
    if ( it == end() )
        set_coeff( var, coeff );
    else {
        it->second += coeff;
        if ( std::fabs(it->second) <= std::numeric_limits<double>::epsilon() ) {
            erase(it);
        }
    }
}


void Constraint::set_coeff( const Variable* var, double coeff )
{
    assert( var != NULL );
    if ( std::fabs(coeff) > std::numeric_limits<double>::epsilon() ) {
        (*this)[var] = coeff;
    }
    else {
        erase(var);
    }
}


std::ostream& operator<<( std::ostream& out, const Constraint* constraint )
{
    assert( constraint != NULL );
    out << constraint->get_name() << ":";
    for ( Constraint::const_iterator it = constraint->begin();
          it != constraint->end();
          ++it )
    {
        double c = it->second;

        if ( c >= 0 )
            out << "  +" << c;
        else
            out << "  -" << -c;
        // out << " x" << it->first->get_index();
        out << " " << it->first->get_name();
    }

    switch ( constraint->get_sense() ) {
        case Constraint::LE: out << " <= "; break;
        case Constraint::EQ: out << " == "; break;
        case Constraint::GE: out << " >= "; break;
    }

    out << constraint->get_rhs();

    return out;
}
