/*
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see  <http://www.gnu.org/licenses/>
 */

#ifndef __CONSTRAINT_HXX__
#define __CONSTRAINT_HXX__

#include <map>
#include <vector>
#include <string>
#include <iosfwd>

class Variable;

/// A constraint.
class Constraint: public std::map<const Variable*, double, bool (*)(const Variable* v1, const Variable* v2)>
{
    typedef std::map<const Variable*, double, bool (*)(const Variable* v1, const Variable* v2)> super_type;
public:
    enum Sense { LE, EQ, GE };

private:
    Sense _sense;
    double _rhs;
    std::string _name;

public:
    Constraint ( Sense sense = EQ, double rhs = 0, const std::string& name = std::string());

    /// Returns the sense of this constraint.
    Sense get_sense() const { return _sense; }

    /// Returns the rhs of this constraint.
    double get_rhs() const { return _rhs; }

    /// Returns the name of this constraint.
    const std::string& get_name() const { return _name; }

    /// Returns all indices of variables (in order).
    std::vector<int> get_indices() const;

    /// Returns the coefficients of all variables (in order).
    std::vector<double> get_coeffs() const;


    /**
     * \brief Adds a coefficient to this constraints.
     *
     * If the variable is already contained in this constraint, it's
     * coefficient is increased by the given value.
     */
    void add_coeff( const Variable* var, double coeff );

    /// Sets the coefficient of a certain variable.
    void set_coeff( const Variable* var, double coeff );

    /// Sets rhs
    void set_rhs( double rhs ) { _rhs = rhs; };
};


std::ostream& operator<<( std::ostream& out, const Constraint* constraint );


#endif
