/*
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see  <http://www.gnu.org/licenses/>
 */

#include "CplexSolver.hxx"

#include "Constraint.hxx"
#include "Variable.hxx"
#include "Problem.hxx"
#include "BaseModel.hxx"
#include "Department.hxx"

#include <iostream>
#include <cstdlib>
#include <cmath>
#include <cassert>

CplexSolver::CplexSolver(double time_limit):
    _lp(0),
    _lower_bound(std::numeric_limits<double>::infinity())
{
    int status;
    _env = CPXopenCPLEX(&status);

    if (_env == NULL) {
        std::cerr << "CPLEX-Error: " << status << std::endl;
        std::exit(EXIT_FAILURE);
    }
    
    //CPXsetintparam( _env, CPX_PARAM_TUNINGDISPLAY, 3 );
    //CPXsetintparam( _env, CPX_PARAM_SCRIND, CPX_ON );

    CPXsetdblparam( _env, CPX_PARAM_EPGAP, 10e-8 );
    if ( time_limit < std::numeric_limits<double>::infinity() )
        CPXsetdblparam( _env, CPX_PARAM_TILIM, time_limit);

    CPXsetintparam( _env, CPX_PARAM_THREADS, 1);
}


CplexSolver::~CplexSolver()
{
    if (_lp) CPXfreeprob(_env, &_lp);
    CPXcloseCPLEX(&_env);
}


double CplexSolver::solve(BaseModel& model)
{
    if ( model.use_cuts() ){
        CPXsetintparam( _env, CPX_PARAM_REDUCE, 0 );
        CPXsetintparam ( _env, CPX_PARAM_PRELINEAR, 0);
        CPXsetintparam ( _env, CPX_PARAM_MIPCBREDLP, CPX_OFF);
        CPXsetusercutcallbackfunc(_env, &cpx_cut_callback, this);
        CPXsetlazyconstraintcallbackfunc(_env, &cpx_cutlazy_callback, this);
    }

    _model = &model;
    _problem = &_model->get_problem();

    int status;
    build_lp();

    CPXsetdblparam( _env, CPX_PARAM_CUTUP, _model->get_upper());

    status = CPXchgobjsen(_env, _lp, CPX_MIN);
    if (status != 0) {
        std::cerr << "Error CPXchgobjsen: " << status << std::endl;
        std::exit(EXIT_FAILURE);
    }
    //CPXwriteprob(_env, _lp, "file.lp", NULL);
    status = CPXmipopt( _env, _lp );
    if (status != 0) {
        std::cerr << "Error CPXmipopt: " << status << std::endl;
        std::exit(EXIT_FAILURE);
    }

    int stat = CPXgetstat( _env, _lp);
    double objval;
    status = CPXgetbestobjval(_env, _lp, &objval);
    _lower_bound = objval;

    if ( status != 0 ) {
        std::cerr << "Error CPXgetbestobjval: " << status << std::endl;
        std::exit(EXIT_FAILURE);
    }

    if (stat != 101 && stat != 102) {
        // not solved to sufficient precision -> infeasible
        return std::numeric_limits<double>::max();
    }

    return objval;
}

void CplexSolver::build_lp()
{
    int status;

    if (_lp) CPXfreeprob(_env, &_lp);
    _lp = CPXcreateprob(_env, &status, "row_layout");
    if (status != 0) {
        std::cerr << "Error CPXcreateprob: " << status << std::endl;
        std::exit(EXIT_FAILURE);
    }

    size_type nvars = _model->get_num_variables();

    std::vector<double> costs(nvars, 0.0);
    std::vector<double> lb( nvars, 0.0 );
    std::vector<double> ub( nvars, std::numeric_limits<double>::max() );
    std::vector<char> types( nvars, 'C' );
    std::vector<std::string> names( nvars );
    std::vector<char*> varnames( nvars );

    for (size_type i = 0; i < nvars; ++i ) {
        const Variable* var = _model->get_variable(i);
        names[i] = var->get_name();
        varnames[i] = const_cast<char*>(names[i].c_str());
        lb[i] = var->get_lb();
        ub[i] = var->get_ub();
        costs[i] = var->get_cost();
        switch (var->get_type()) {
            case Variable::Cont:
                types[i] = 'C';
                break;
            case Variable::Binary:
                types[i] = 'B';
                break;
        }
    }

    status = CPXnewcols( _env, _lp, static_cast<int>(nvars), &(costs[0]), &(lb[0]), &(ub[0]), &(types[0]), &(varnames[0]) );
    if ( status != 0 ) {
        std::cerr << "Error CPXnewcols - mip-vars: " << status << std::endl;
        std::exit(EXIT_FAILURE);
    }

    // add constraints

    for ( size_type i = 0; i < _model->get_num_constraints(); ++i ) {
        const Constraint* constraint = _model->get_constraint(i);

        std::vector<int> indices = constraint->get_indices();
        std::vector<double> coeffs = constraint->get_coeffs();
        double rhs[1];
        rhs[0] = constraint->get_rhs();

        char sense[1];
        switch (constraint->get_sense()) {
            case Constraint::LE: sense[0] = 'L'; break;
            case Constraint::EQ: sense[0] = 'E'; break;
            case Constraint::GE: sense[0] = 'G'; break;
            default: assert(0);
        }

        int rmatbeg[1] = { 0 };

        std::string name = constraint->get_name();
        char* name_array[1];
        char** row_name = NULL;

        if (!name.empty()) {
            name_array[0] = &(name[0]);
            row_name = name_array;
        }
        status = CPXaddrows(_env, _lp,
                            0, 1, 
                            static_cast<int>(constraint->size()),
                            rhs,
                            sense,
                            rmatbeg, &(indices[0]), &(coeffs[0]),
                            NULL, row_name);
        if ( status != 0 ) {
            std::cerr << "Error CPXnewrows: " << status << std::endl;
            std::exit(EXIT_FAILURE);
        }
    }
}

std::vector<double> CplexSolver::get_solution() const
{
    std::vector<double> sol( CPXgetnumcols(_env, _lp) );
    int status;

    status = CPXgetx( _env, _lp, &(sol[0]), 0, static_cast<int>(sol.size()) - 1 );
    if ( status != 0 ) {
        std::cerr << "Error CPXgetx - lp: " << status << std::endl;
        std::exit(EXIT_FAILURE);
    }

    return sol;
}

int CplexSolver::cpx_cut_callback(CPXCENVptr env, void* cbdata, int wherefrom, void* cbhandle, int* useraction_p)
{
    return reinterpret_cast<CplexSolver*>(cbhandle)->cut_callback(env, cbdata, wherefrom, useraction_p);
}

int CplexSolver::cpx_cutlazy_callback(CPXCENVptr env, void* cbdata, int wherefrom, void* cbhandle, int* useraction_p)
{
    return reinterpret_cast<CplexSolver*>(cbhandle)->cutlazy_callback(env, cbdata, wherefrom, useraction_p);
}


int CplexSolver::cut_callback(CPXCENVptr env, void* cbdata, int wherefrom, int* useraction_p)
{
    int status;
    _cbdata = cbdata;
    _wherefrom = wherefrom;

    _x.resize( CPXgetnumcols(env, _lp) );
    status = CPXgetcallbacknodex( env, cbdata, wherefrom,
                                  &(_x[0]), 0, static_cast<int>(_x.size()) - 1 );
    if ( status != 0 ) {
        std::cerr << "Error CPXgetcallbacknodex: " << status << std::endl;
        std::abort();
    }

    _model->separate(_x, *this);

    *useraction_p = CPX_CALLBACK_DEFAULT;
    return 0;
}

int CplexSolver::cutlazy_callback(CPXCENVptr env, void* cbdata, int wherefrom, int* useraction_p)
{
    int status;
    _cbdata = cbdata;
    _wherefrom = wherefrom;

    _x.resize( CPXgetnumcols(env, _lp) );
    status = CPXgetcallbacknodex( env, cbdata, wherefrom,
                                  &(_x[0]), 0, static_cast<int>(_x.size()) - 1 );
    if ( status != 0 ) {
        std::cerr << "Error CPXgetcallbacknodex: " << status << std::endl;
        std::abort();
    }

    _model->separate(_x, *this);

    *useraction_p = CPX_CALLBACK_DEFAULT;
    return 0;
}

bool CplexSolver::add_if_violated(const coeff_vec& variables,
                                  Constraint::Sense sense,
                                  double rhs )
{
    double sum = 0.0;
    for ( size_type i = 0; i < variables.size(); ++i ) {
        sum += _x[variables[i].first->get_index()] * variables[i].second;
    }

    bool violated = false;
    char csense;
    switch (sense) {
        case Constraint::LE:
            violated = sum - rhs > 1e-4;
            csense = 'L';
            break;
        case Constraint::GE:
            violated = sum - rhs < -1e-4;
            csense = 'G';
            break;
        case Constraint::EQ:
            violated = std::fabs(sum - rhs) > 1e-4;
            csense = 'E';
            break;
    }

    if ( violated ) {
        std::vector<int> inds;
        std::vector<double> vals;
        for ( size_type i = 0; i < variables.size(); ++i ) {
            inds.push_back((int)variables[i].first->get_index());
            vals.push_back(variables[i].second);
        }

        int status;

        status = CPXcutcallbackadd(_env, _cbdata, _wherefrom,
                                   (int)variables.size(), rhs, csense,
                                   &(inds[0]), &(vals[0]), CPX_USECUT_FORCE);

        if ( status != 0 ) {
            std::cerr << "CPXcutcallbackadd: " << status << std::endl;
            std::abort();
        }
    }
    return violated;
}
