/*
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see  <http://www.gnu.org/licenses/>
 */

#include "DRFLPBetweennessModel.hxx"

#include "Problem.hxx"
#include "Department.hxx"
#include "Constraint.hxx"
#include "Variable.hxx"
#include "InitSeparator.hxx"

DRFLPBetweennessModel::DRFLPBetweennessModel(Problem& problem,
                                             const rows_vec& rows,
                                             bool use_cuts,
                                             double bigm,
                                             double upper):
    BetweennessModel(problem, rows, use_cuts, upper), _bigm(bigm)
{
}


void DRFLPBetweennessModel::init()
{
    init_x_variables();
    init_d_variables(_problem.get_num_orig() + 1);
    init_betweenness_constraints();
    init_drflp_constraints();
}


void DRFLPBetweennessModel::init_drflp_constraints()
{
    size_type n = _problem.get_num_orig();

    for ( size_type m = 0; m < _rows.size(); m++ ){
        for (size_type i = 0; i < _rows[m].size(); i++){
            for (size_type j = 0; j < _rows[m].size(); j++){
                if ( i == j ) continue;
                size_type idx_i, idx_j;
                idx_i = _rows[m][i];
                idx_j = _rows[m][j];
                double li = department(idx_i)->get_length();
                double lj = department(idx_j)->get_length();
                Constraint* constraint = new Constraint( Constraint::GE, -_bigm + li*0.5 + lj*0.5);
                constraint->set_coeff( _d_variables[idx_i][n], -1.0 );
                constraint->set_coeff( _d_variables[idx_j][n], 1.0 );
                constraint->set_coeff( _x_variables[n][idx_i][idx_j], -_bigm );

                _constraints.push_back(constraint);
            }
        }
    }

    for ( size_type m = 0; m < _rows.size(); m++ ){
        for ( size_type i = 0; i < _rows[m].size()+1; i++){
            for ( size_type j = i+1; j < _rows[m].size()+1; j++){
                size_type idx_i, idx_j;
                if ( i < _rows[m].size() ) idx_i = _rows[m][i];
                if ( j < _rows[m].size() ) idx_j = _rows[m][j];
                if ( i == _rows[m].size() ) idx_i = n;
                if ( j == _rows[m].size() ) idx_j = n;
                double li = department(idx_i)->get_length();
                double lj = department(idx_j)->get_length();

                Constraint* constraint = new Constraint( Constraint::GE, (li+lj)*0.5);
                constraint->set_coeff( _d_variables[idx_i][idx_j], 1.0 );
                for ( size_type k = 0; k < _rows[m].size(); k++ ){
                    if ( k == i || k == j ) continue;
                    double lk = department(_rows[m][k])->get_length();
                    constraint->set_coeff( _x_variables[idx_i][_rows[m][k]][idx_j], -lk );
                }
                _constraints.push_back(constraint);
            }
        }
    }

    if ( use_cuts() == false ){
        InitSeparator isep(_constraints);
        separate_triangle_constraints(isep);
    }

    for ( size_type i = 0; i < n; i++){
        Constraint* constraint = new Constraint( Constraint::LE, _bigm);
        constraint->set_coeff( _d_variables[i][n], 1.0 );
        _constraints.push_back(constraint);
    }
    // symmetry
    {
		bool found = false;
		for ( size_type m = 0; m < _rows.size(); m++ ){

        	if ( _rows[m].size() > 1 && found == false ){
            	Constraint* constraint = new Constraint( Constraint::EQ, 1.0);
            	constraint->set_coeff( _x_variables[n][_rows[m][0]][_rows[m][1]], 1.0 );
            	_constraints.push_back(constraint);
   				found = true;
        	}
		}
    }
}


void DRFLPBetweennessModel::separate(const std::vector<double>& /*x*/, Separator& separator)
{
    if (use_cuts()) {
        separate_betweenness_constraints(separator);
        separate_triangle_constraints(separator);
    }
}
