/*
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see  <http://www.gnu.org/licenses/>
 */

#ifndef __LYUTIL_ERROR_HXX__
#define __LYUTIL_ERROR_HXX__

#include <stdexcept>
#include <cassert>
#include <iostream>
#include <sstream>
#include <cstdarg>

namespace lyutil {

struct RuntimeError: public std::runtime_error {
protected:
    explicit RuntimeError(const std::string& name);

    explicit RuntimeError(const std::string& name,
                          const std::string& msg);

    explicit RuntimeError(const std::string& name,
                          const std::string& where,
                          const std::string& msg);
};

void throw_error();

#define LYUTIL_DEF_ERROR(name)	  \
    struct name: public lyutil::RuntimeError { \
        explicit name(): lyutil::RuntimeError(#name) {} \
        explicit name(const std::string& msg): \
            lyutil::RuntimeError(#name, msg) {} \
        explicit name(const std::string& where, \
                      const std::string& msg): \
            lyutil::RuntimeError(#name, where, msg) {} \
    };


LYUTIL_DEF_ERROR(FileError)
LYUTIL_DEF_ERROR(UnsupportedError)

#include <iostream>
#ifdef USE_STDERR
#define LYUTIL_TEST(expression, e, where, args)	  \
    if (!(expression)){ \
        std::cerr << "Check failed in file " << __FILE__ << " at line " << __LINE__ << ":" << std::endl; \
        std::cerr << where << std::endl; \
        std::cerr << #expression << std::endl; \
        std::ostringstream o; \
        o << args; \
        lyutil::throw_error(); \
        throw(e ((std::string("(")  + where + "): " + o.str()).c_str())); \
    }
#else
#define LYUTIL_TEST(expression, e, where, args)	  \
    if (!(expression)) { \
        std::ostringstream o; \
        o << args; \
        lyutil::throw_error(); \
        throw(e ((std::string("(")  + where + "): " + o.str()).c_str())); \
    }
#endif

#if !(defined NDEBUG) || (defined LYUTIL_FORCE_ASSERT)
#define LYUTIL_ASSERT(expression, e, where, args)	  \
    LYUTIL_TEST(expression, e, where, args)

#else
#define LYUTIL_ASSERT(expression, e, where, args)	  \
    if ( false || (expression) ) {};
#endif

}

#define LYUTIL_TEST2( expression, e, args )	  \
    LYUTIL_TEST( expression, e, __PRETTY_FUNCTION__, args )

#define LYUTIL_ASSERT2( expression, e, args )	  \
    LYUTIL_ASSERT( expression, e, __PRETTY_FUNCTION__, args )

#endif
