/*
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see  <http://www.gnu.org/licenses/>
 */

#include "GurobiSolver.hxx"

#include "Constraint.hxx"
#include "Variable.hxx"
#include "Problem.hxx"
#include "BaseModel.hxx"
#include "Department.hxx"

#include <cmath>

GurobiSolver::GurobiSolver(double time_limit):
    _env(0), _lp(0), _vars(0),
    _lower_bound(std::numeric_limits<double>::infinity())
{
    _env = new GRBEnv();
    _env->set(GRB_IntParam_Threads, 1);
    _env->set(GRB_IntParam_LogToConsole, 0);
    _env->set(GRB_DoubleParam_MIPGap, 1e-8);
    _env->set(GRB_StringParam_LogFile, "rowprob.grblog");
    if (time_limit < std::numeric_limits<double>::infinity())
        _env->set(GRB_DoubleParam_TimeLimit, time_limit);
}


GurobiSolver::~GurobiSolver()
{
    delete[] _vars;
    delete _lp;
    delete _env;
}


double GurobiSolver::solve(BaseModel& model)
{
    try {
        if ( model.use_cuts() ){
            /// set some parameter?
            _env->set(GRB_IntParam_PreCrush, 1);
            _env->set(GRB_IntParam_LazyConstraints, 1);
            _env->set(GRB_IntParam_DualReductions, 0);
        }

        _model = &model;
        _problem = &_model->get_problem();
        _env->set(GRB_DoubleParam_Cutoff, _model->get_upper());

        build_lp();

        _lp->optimize();

        if (_lp->get(GRB_IntAttr_Status) == GRB_OPTIMAL) {
            _lower_bound = _lp->get(GRB_DoubleAttr_ObjVal);
            return _lower_bound;
        } else {
            _lower_bound = _lp->get(GRB_DoubleAttr_ObjBound);
            return std::numeric_limits<double>::infinity();
        }
    }
    catch (GRBException& e) {
        std::cerr << "Gurobi ERROR (" << e.getErrorCode() << "): "
                  << e.getMessage() << std::endl;
        throw;
    }
}

void GurobiSolver::build_lp()
{
    delete _lp;

    _lp = new GRBModel(*_env);
    _lp->setCallback(this);
    _lp->set(GRB_IntAttr_ModelSense, +1);

    size_type nvars = _model->get_num_variables();

    std::vector<double> costs(nvars, 0.0);
    std::vector<double> lb( nvars, 0.0 );
    std::vector<double> ub( nvars, std::numeric_limits<double>::max() );
    std::vector<char> types( nvars, 'C' );
    std::vector<std::string> names( nvars );

    for (size_type i = 0; i < nvars; ++i ) {
        const Variable* var = _model->get_variable(i);
        names[i] = var->get_name();
        lb[i] = var->get_lb();
        ub[i] = var->get_ub();
        costs[i] = var->get_cost();
        switch (var->get_type()) {
            case Variable::Cont: types[i] = GRB_CONTINUOUS; break;
            case Variable::Binary: types[i] = GRB_BINARY; break;
        }
    }

    delete[] _vars;
    _vars = _lp->addVars(&(lb[0]), &(ub[0]), &(costs[0]), &(types[0]), &(names[0]), nvars);

    _lp->update();

    // add constraints
    for ( size_type i = 0; i < _model->get_num_constraints(); ++i ) {
        const Constraint* c = _model->get_constraint(i);

        GRBLinExpr expr = 0;
        for (Constraint::const_iterator it = c->begin(); it != c->end(); ++it) {
            expr += _vars[it->first->get_index()] * it->second;
        }

        GRBTempConstr constr;
        switch (c->get_sense()) {
            case Constraint::LE: constr = expr <= c->get_rhs(); break;
            case Constraint::EQ: constr = expr == c->get_rhs(); break;
            case Constraint::GE: constr = expr >= c->get_rhs(); break;
        }

        _lp->addConstr(constr);
        // _lp->addConstr(constr, c->get_name());
    }

    _lp->update();
}


std::vector<double> GurobiSolver::get_solution() const
{
    int nvars = _lp->get(GRB_IntAttr_NumVars);
    double* data = _lp->get(GRB_DoubleAttr_X, _vars, nvars);
    std::vector<double> sol(data, data + nvars);
    delete[] data;
    return sol;
}


void GurobiSolver::callback()
{
    try {
        if ( GRBCallback::where != GRB_CB_MIPSOL &&
             GRBCallback::where != GRB_CB_MIPNODE )
            return;

        int nvars = _lp->get(GRB_IntAttr_NumVars);

        double* xdata;
        if (GRBCallback::where == GRB_CB_MIPSOL) {
            xdata = GRBCallback::getSolution(_vars, nvars);
        } else {
            if ( GRBCallback::getIntInfo(GRB_CB_MIPNODE_STATUS) != GRB_OPTIMAL )
                return;
            xdata = GRBCallback::getNodeRel(_vars, nvars);
        }

        _x.assign(xdata, xdata + nvars);
        delete[] xdata;

        _model->separate(_x, *this);
    }
    catch (GRBException& e) {
        std::cerr << e.getMessage() << std::endl;
        std::cerr << e.getErrorCode() << std::endl;
        throw;
    }
}

bool GurobiSolver::add_if_violated(const coeff_vec& variables,
                                   Constraint::Sense sense,
                                   double rhs )
{
    double sum = 0.0;
    for ( size_type i = 0; i < variables.size(); ++i ) {
        sum += _x[variables[i].first->get_index()] * variables[i].second;
    }

    bool violated = false;
    switch (sense) {
        case Constraint::LE:
            violated = sum - rhs > 1e-4;
            break;
        case Constraint::GE:
            violated = sum - rhs < -1e-4;
            break;
        case Constraint::EQ:
            violated = std::fabs(sum - rhs) > 1e-4;
            break;
    }

    if ( violated ) {
        GRBLinExpr expr;
        for ( size_type i = 0; i < variables.size(); ++i ) {
            expr += _vars[variables[i].first->get_index()] * variables[i].second;
        }

        GRBTempConstr constr;
        switch (sense) {
            case Constraint::LE: constr = expr <= rhs; break;
            case Constraint::EQ: constr = expr == rhs; break;
            case Constraint::GE: constr = expr >= rhs; break;
        }

        if (GRBCallback::where == GRB_CB_MIPSOL)
            GRBCallback::addLazy(constr);
        else
            GRBCallback::addCut(constr);
    }

    return violated;
}
