/*
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see  <http://www.gnu.org/licenses/>
 */

#include "PROPBetweennessModel.hxx"

#include "Problem.hxx"
#include "Department.hxx"
#include "Constraint.hxx"
#include "Variable.hxx"
#include "InitSeparator.hxx"

#include <cmath>

PROPBetweennessModel::PROPBetweennessModel(Problem& problem,
                                           const rows_vec& rows,
                                           bool use_cuts,
                                           bool use_cliques,
                                           CostType ctype,
                                           double upper):
    BetweennessModel(problem, rows, use_cuts, upper), _ctype(ctype), _use_cliques(use_cliques)
{
}


void PROPBetweennessModel::init()
{
    init_x_variables();

    if (_ctype == Cost2 )
        init_cost2();
    else
        init_d_variables(_problem.get_num_orig() + 1);

    init_betweenness_constraints();
    init_prop_constraints();
}


void PROPBetweennessModel::init_cost2()
{
    size_type n = _problem.get_num_orig();
    for (size_type d1 = 0; d1 < _x_variables.size(); d1++) {
        for (size_type d2 = 0; d2 < _x_variables[d1].size(); d2++) {
            if ( d1 == d2 ) continue;
            for (size_type d3 = 0; d3 < _x_variables[d1][d2].size(); d3++) {
                if ( d3 == d2 || d3 == d1 ) continue;
                Variable* v = _x_variables[d1][d2][d3];
                if ( v == NULL ) continue;
                const Department* dep1 = v->get_dep1();
                const Department* dep2 = v->get_dep2();
                const Department* dep3 = v->get_dep3();

                if ( dep1->get_id() < n &&
                     dep2->get_id() < n &&
                     dep3->get_id() < n )
                {
                    v->set_cost(get_transport(dep1, dep3) * dep2->get_length());
                }
                else if ( dep3->get_id() == n &&
                          dep2->get_id() < n &&
                          dep1->get_id() < n)
                {
                    size_type r1 = dep1->get_row();
                    double c = 0.0;
                    for ( size_type j = 0; j < _rows.size(); j++){
                        if ( j == r1 ) continue;
                        for ( size_type k = 0; k < _rows[j].size(); k++){
                            size_type m2 = _rows[j][k];
                            c += get_transport(dep1, department(m2));
                        }
                    }

                    v->set_cost(dep2->get_length() * c);
                }
            }
        }
    }
}

void PROPBetweennessModel::init_prop_constraints()
{
    size_type n = _problem.get_num_orig();
    if ( _ctype == Cost1 ) {
        for ( size_type l = 0; l < _rows.size(); l++ ) {
            for ( size_type o = l+1; o < _rows.size(); o++) {
                for (size_type i = 0; i < _rows[l].size(); i++) {
                    for (size_type j = 0; j < _rows[o].size(); j++) {
                        double li = department(_rows[l][i])->get_length();
                        double lj = department(_rows[o][j])->get_length();
                        Constraint* constraint = new Constraint( Constraint::GE, 0.5*(lj-li));
                        constraint->set_coeff( _d_variables[_rows[l][i]][_rows[o][j]], 1.0 );
                        for ( size_type k = 0; k < _rows[l].size(); k++ ) {
                            if ( k == i ) continue;
                            double lk = department(_rows[l][k])->get_length();
                            constraint->set_coeff( _x_variables[n][_rows[l][k]][_rows[l][i]], lk );
                        }
                        for ( size_type k = 0; k < _rows[o].size(); k++ ) {
                            if ( k == j ) continue;
                            double lk = department(_rows[o][k])->get_length();
                            constraint->set_coeff( _x_variables[n][_rows[o][k]][_rows[o][j]], -lk );
                        }
                        _constraints.push_back(constraint);
                    }
                }
            }
        }

        for ( size_type l = 0; l < _rows.size(); l++ ) {
            for ( size_type o = l+1; o < _rows.size(); o++) {
                for (size_type i = 0; i < _rows[l].size(); i++) {
                    for (size_type j = 0; j < _rows[o].size(); j++) {
                        double li = department(_rows[l][i])->get_length();
                        double lj = department(_rows[o][j])->get_length();
                        Constraint* constraint = new Constraint( Constraint::GE, 0.5*(li-lj));
                        constraint->set_coeff( _d_variables[_rows[l][i]][_rows[o][j]], 1.0 );
                        for ( size_type k = 0; k < _rows[l].size(); k++ ) {
                            if ( k == i ) continue;
                            double lk = department(_rows[l][k])->get_length();
                            constraint->set_coeff( _x_variables[n][_rows[l][k]][_rows[l][i]], -lk );
                        }
                        for ( size_type k = 0; k < _rows[o].size(); k++ ) {
                            if ( k == j ) continue;
                            double lk = department(_rows[o][k])->get_length();
                            constraint->set_coeff( _x_variables[n][_rows[o][k]][_rows[o][j]], lk );
                        }
                        _constraints.push_back(constraint);
                    }
                }
            }
        }
    }

    if ( _ctype == Cost1 || _ctype == Cost3 ) {
        for ( size_type m = 0; m < _rows.size(); m++ ) {
            for ( size_type i = 0; i < _rows[m].size()+1; i++) {
                for ( size_type j = i+1; j < _rows[m].size()+1; j++) {
                    size_type idx_i, idx_j;
                    if ( i < _rows[m].size() ) idx_i = _rows[m][i];
                    if ( j < _rows[m].size() ) idx_j = _rows[m][j];
                    if ( i == _rows[m].size() ) idx_i = n;
                    if ( j == _rows[m].size() ) idx_j = n;
                    double li = department(idx_i)->get_length();
                    double lj = department(idx_j)->get_length();
                    Constraint* constraint = new Constraint( Constraint::EQ, 0.5*(li+lj));
                    constraint->set_coeff( _d_variables[idx_i][idx_j], 1.0 );
                    for ( size_type k = 0; k < _rows[m].size(); k++ ) {
                        if ( k == i || k == j ) continue;
                        double lk = department(_rows[m][k])->get_length();
                        constraint->set_coeff( _x_variables[idx_i][_rows[m][k]][idx_j], -lk );
                    }
                    _constraints.push_back(constraint);
                }
            }
        }
    }

    if ( use_cuts() == false ) {
        if ( _ctype == Cost1 ) {
            InitSeparator isep(_constraints);
            separate_triangle_constraints(isep);
        }
        else if ( _ctype == Cost3 ) {
            InitSeparator isep(_constraints);
            separate_triangle_constraints(isep, 2);
        }
    }

    if (_ctype == Cost3 ) {
        for ( size_type l = 0; l < _rows.size(); l++ ) {
            for ( size_type o = l+1; o < _rows.size() && o-1 == l ; o++) {
                for (size_type i = 0; i < _rows[l].size(); i++) {
                    for (size_type j = 0; j < _rows[o].size(); j++) {
                        double li = department(_rows[l][i])->get_length();
                        double lj = department(_rows[o][j])->get_length();
                        Constraint* constraint = new Constraint( Constraint::GE, 0.5*(lj-li));
                        constraint->set_coeff( _d_variables[_rows[l][i]][_rows[o][j]], 1.0 );
                        for ( size_type k = 0; k < _rows[l].size(); k++ ) {
                            if ( k == i ) continue;
                            double lk = department(_rows[l][k])->get_length();
                            constraint->set_coeff( _x_variables[n][_rows[l][k]][_rows[l][i]], lk );
                        }
                        for ( size_type k = 0; k < _rows[o].size(); k++ ) {
                            if ( k == j ) continue;
                            double lk = department(_rows[o][k])->get_length();
                            constraint->set_coeff( _x_variables[n][_rows[o][k]][_rows[o][j]], -lk );
                        }
                        _constraints.push_back(constraint);
                    }
                }
            }
        }
        for ( size_type l = 0; l < _rows.size(); l++ ) {
            for ( size_type o = l+1; o < _rows.size() && o - 1 == l; o++) {
                for (size_type i = 0; i < _rows[l].size(); i++) {
                    for (size_type j = 0; j < _rows[o].size(); j++) {
                        double li = department(_rows[l][i])->get_length();
                        double lj = department(_rows[o][j])->get_length();
                        Constraint* constraint = new Constraint( Constraint::GE, 0.5*(li-lj));
                        constraint->set_coeff( _d_variables[_rows[l][i]][_rows[o][j]], 1.0 );
                        for ( size_type k = 0; k < _rows[l].size(); k++ ) {
                            if ( k == i ) continue;
                            double lk = department(_rows[l][k])->get_length();
                            constraint->set_coeff( _x_variables[n][_rows[l][k]][_rows[l][i]], -lk );
                        }
                        for ( size_type k = 0; k < _rows[o].size(); k++ ) {
                            if ( k == j ) continue;
                            double lk = department(_rows[o][k])->get_length();
                            constraint->set_coeff( _x_variables[n][_rows[o][k]][_rows[o][j]], lk );
                        }
                        _constraints.push_back(constraint);
                    }
                }
            }
        }
        for ( size_type l = 0; l < _rows.size(); l++ ) {
            for ( size_type o = l+2; o < _rows.size() ; o++) {
                for (size_type i = 0; i < _rows[l].size(); i++) {
                    for (size_type j = 0; j < _rows[o].size(); j++) {
                        double li = department(_rows[l][i])->get_length();
                        double lj = department(_rows[o][j])->get_length();
                        Constraint* constraint = new Constraint( Constraint::EQ, 0.5*(lj+li));
                        constraint->set_coeff( _d_variables[_rows[l][i]][_rows[o][j]], 1.0 );
                        for ( size_type k = 0; k < _rows[l].size(); k++ ) {
                            if ( k == i ) continue;
                            double lk = department(_rows[l][k])->get_length();
                            constraint->set_coeff( _x_variables[n][_rows[l][k]][_rows[l][i]], -lk );
                        }
                        for ( size_type k = 0; k < _rows[o].size(); k++ ) {
                            if ( k == j ) continue;
                            double lk = department(_rows[o][k])->get_length();
                            constraint->set_coeff( _x_variables[n][_rows[o][k]][_rows[o][j]], -lk );
                        }
                        _constraints.push_back(constraint);
                    }
                }
            }
        }
    }

    if ( _use_cliques ) {

        for ( size_type l = 0; l < _rows.size(); l++ ) {
            double rhs = 0.;
            for ( size_type i = 0; i < _rows[l].size(); i++ ) {
                rhs = rhs + department(_rows[l][i])->get_length();
            }
            rhs = rhs*rhs*rhs/6.0;
            for ( size_type i = 0; i < _rows[l].size(); i++ ) {
                rhs = rhs - department(_rows[l][i])->get_length()*
                    department(_rows[l][i])->get_length()*
                    department(_rows[l][i])->get_length()/6.0;
            }
            Constraint* constraint = new Constraint( Constraint::EQ, rhs);
            for (size_type i = 0; i < _rows[l].size(); i++) {
                for (size_type j = i+1; j < _rows[l].size(); j++) {
                    double li = department(_rows[l][i])->get_length();
                    double lj = department(_rows[l][j])->get_length();
                    constraint->set_coeff( _d_variables[_rows[l][i]][_rows[l][j]], li*lj );
                }
            }
            _constraints.push_back(constraint);
        }

        if ( use_cuts() == false ) {
            InitSeparator isep(_constraints);
            separate_clique_constraints(isep);
        }
    }
}

double PROPBetweennessModel::compute_fixed_costs()
{
    if ( _ctype != Cost2 ) return 0;

    double fixcosts = 0.0;
    for ( size_type i = 0; i < _rows.size(); i++ ){
        for ( size_type j = 0; j < _rows[i].size(); j++){
            for ( size_type k = j+1; k < _rows[i].size(); k++){
                fixcosts += 0.5
                    * get_transport(department(_rows[i][j]), department(_rows[i][k]))
                    * (department(_rows[i][j])->get_length() + department(_rows[i][k])->get_length() );
            }
        }
        for ( size_type i2 = i+1; i2 < _rows.size(); i2++ ){
            for ( size_type j = 0; j < _rows[i].size(); j++){
                for ( size_type k = 0; k < _rows[i2].size(); k++){
                    fixcosts +=
                        0.5
                        * get_transport(department(_rows[i][j]), department(_rows[i2][k]))
                        * (department(_rows[i][j])->get_length() + department(_rows[i2][k])->get_length() );
                }
            }
        }

    }

    return fixcosts;
}

void PROPBetweennessModel::separate_clique_constraints(Separator& separator)
{
    for ( size_type l = 0; l < _rows.size(); l++ ){
        for ( size_type i = 0; i < _rows[l].size(); i++ ){
            for ( size_type j = i+1; j < _rows[l].size(); j++ ){
                for ( size_type k = j+1; k < _rows[l].size(); k++ ){
                    double leni = department(_rows[l][i])->get_length();
                    double lenj = department(_rows[l][j])->get_length();
                    double lenk = department(_rows[l][k])->get_length();

                    double rhs = leni + lenj + lenk;
                    rhs = rhs*rhs*rhs/6.0;
                    rhs -= leni*leni*leni/6.0;
                    rhs -= lenj*lenj*lenj/6.0;
                    rhs -= lenk*lenk*lenk/6.0;

                    Separator::coeff_vec vars;
                    vars.push_back( std::make_pair( _d_variables[_rows[l][i]][_rows[l][j]], leni * lenj) );
                    vars.push_back( std::make_pair( _d_variables[_rows[l][i]][_rows[l][k]], leni * lenk) );
                    vars.push_back( std::make_pair( _d_variables[_rows[l][j]][_rows[l][k]], lenj * lenk) );
                    separator.add_if_violated( vars, Constraint::GE, rhs);
                }
            }
        }
    }
}

void PROPBetweennessModel::separate(const std::vector<double>& /*x*/, Separator& separator)
{
    if (use_cuts()) {
        separate_betweenness_constraints(separator);

        if (_ctype == Cost1)
            separate_triangle_constraints(separator);
        else if (_ctype == Cost3)
            separate_triangle_constraints(separator, 2);

        if (_use_cliques)
            separate_clique_constraints(separator);
    }
}
