/*
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see  <http://www.gnu.org/licenses/>
 */

#include "SAModelCommon.hxx"

#include "Department.hxx"
#include "InitSeparator.hxx"
#include "Problem.hxx"
#include "Constraint.hxx"
#include "Variable.hxx"

SAModelCommon::SAModelCommon()
{
}

void SAModelCommon::init_e_variables()
{
    size_type n = _problem.get_num_orig();
    _e_variables.resize( n );
    for ( size_type i = 0; i < n; ++i ) {
        _e_variables[i].resize( n );
    }
    for ( size_type i = 0; i < n; ++i ){
        for ( size_type j = 0; j < n; ++j ){
            _e_variables[i][j].assign( n, NULL );
        }
    }
    for ( size_type i = 0; i < n; ++i ) {
        for ( size_type j = 0; j < n; ++j ) {
            for ( size_type k = i+1; k < n; ++k ){
                if ( j == i || j == k ) continue;
                Variable* var = new Variable(department(i),
                                            department(j),
                                            department(k),
                                            _variables.size(),
                                            "e", Variable::Binary);
                _e_variables[i][j][k] = var;
                _e_variables[k][j][i] = var;
                _variables.push_back(var);
            }
        }
    }
}

void SAModelCommon::init_e_constraints()
{
	if ( use_cuts() == false ) {
		InitSeparator isep(_constraints);
		// separate_e_constraints(isep);
	}
}

void SAModelCommon::separate_e_constraints(Separator& separator)
{
    // constraints tested in the paper of Secchin & Amaral (not used here)
    size_type n = _problem.get_num_orig();
    for ( size_type i = 0; i < n; i++){
        for ( size_type j = i+1; j < n; j++){
            for ( size_type k = 0; k < n; k++){
                if ( k == j || k == i ) continue;
                Separator::coeff_vec vars;
                vars.push_back( std::make_pair( _e_variables[i][k][j], 1.0) );
                vars.push_back( std::make_pair( _a_variables[i][k], -1.0) );
                vars.push_back( std::make_pair( _a_variables[k][i], -1.0) );
                separator.add_if_violated( vars, Constraint::LE, 0.0);
                vars.clear();
                vars.push_back( std::make_pair( _e_variables[i][k][j], 1.0) );
                vars.push_back( std::make_pair( _a_variables[i][j], -1.0) );
                vars.push_back( std::make_pair( _a_variables[j][i], -1.0) );
                separator.add_if_violated( vars, Constraint::LE, 0.0);
                vars.clear();
                vars.push_back( std::make_pair( _e_variables[i][k][j], 1.0) );
                vars.push_back( std::make_pair( _a_variables[j][k], -1.0) );
                vars.push_back( std::make_pair( _a_variables[k][j], -1.0) );
                separator.add_if_violated( vars, Constraint::LE, 0.0);
                vars.clear();
            }
        }
    }
}

void SAModelCommon::init_sa_base_constraints()
{
    size_type n = _problem.get_num_orig();
    for ( size_type i = 0; i < n; i++ ){
        for ( size_type j = i+1; j < n; j++){
            Constraint* constraint = new Constraint( Constraint::GE, 0.0);
            constraint->set_coeff( _d_variables[i][j], 1.0 );
            constraint->set_coeff( _a_variables[i][j], -department(i)->get_length()/2.0-department(j)->get_length()/2.0 );
            constraint->set_coeff( _a_variables[j][i], -department(i)->get_length()/2.0-department(j)->get_length()/2.0 );
            for ( size_type k = 0; k < n; k++){
                if ( k == i || k == j ) continue;
                constraint->set_coeff( _e_variables[i][k][j], -department(k)->get_length() );
            }
            _constraints.push_back(constraint);
        }
    }
    for ( size_type i = 0; i < n; i++ ){
        for ( size_type j = i+1; j < n; j++){
            for ( size_type k = 0; k < n; k++){
                if ( k == i || k == j ) continue;
                Constraint* constraint = new Constraint( Constraint::GE, -1.0);
                constraint->set_coeff( _e_variables[i][k][j], 1.0 );
                constraint->set_coeff( _a_variables[i][k], -1.0 );
                constraint->set_coeff( _a_variables[k][j], -1.0 );
                _constraints.push_back(constraint);
            }
        }
    }
    for ( size_type i = 0; i < n; i++ ){
        for ( size_type j = i+1; j < n; j++){
            for ( size_type k = 0; k < n; k++){
                if ( k == i || k == j ) continue;
                Constraint* constraint = new Constraint( Constraint::GE, -1.0);
                constraint->set_coeff( _e_variables[i][k][j], 1.0 );
                constraint->set_coeff( _a_variables[j][k], -1.0 );
                constraint->set_coeff( _a_variables[k][i], -1.0 );
                _constraints.push_back(constraint);
            }
        }
    }

    for ( size_type i = 0; i < n; i++ ){
        for ( size_type j = i+1; j < n; j++){
            for ( size_type k = j+1; k < n; k++){
                {
                    Constraint* constraint = new Constraint( Constraint::LE, 0.0);
                    constraint->set_coeff( _d_variables[i][j], 1.0 );
                    constraint->set_coeff( _d_variables[i][k], -1.0 );
                    constraint->set_coeff( _d_variables[j][k], -1.0 );
                    _constraints.push_back(constraint);
                }
                {
                    Constraint* constraint = new Constraint( Constraint::LE, 0.0);
                    constraint->set_coeff( _d_variables[i][k], 1.0 );
                    constraint->set_coeff( _d_variables[i][j], -1.0 );
                    constraint->set_coeff( _d_variables[j][k], -1.0 );
                    _constraints.push_back(constraint);
                }
                {
                    Constraint* constraint = new Constraint( Constraint::LE, 0.0);
                    constraint->set_coeff( _d_variables[j][k], 1.0 );
                    constraint->set_coeff( _d_variables[i][j], -1.0 );
                    constraint->set_coeff( _d_variables[i][k], -1.0 );
                    _constraints.push_back(constraint);
                }
            }
        }
    }
}



void SAModelCommon::separate(const std::vector<double>& /*x*/,
                             Separator& separator)
{
    // separate_e_constraints(separator);
}
