/*
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see  <http://www.gnu.org/licenses/>
 */

#include "Timer.hxx"

#include <sstream>
#include <ostream>
#include <iomanip>

#ifdef __unix
#include <sys/times.h>
#include <unistd.h>
#endif


namespace lyutil {

#ifdef __unix
const clock_t Timer::SC_CLK_TCK = sysconf(_SC_CLK_TCK);
#else
const clock_t Timer::SC_CLK_TCK = CLOCKS_PER_SEC;
#endif

Timer::
Timer(bool autostart) :
    _running(false),
    _started(false)
{
    if (autostart)
        start();
}


Timer::~Timer()
{
    if (_running) stop();
}


void
Timer::
start(bool reset)
{
    if (reset || !_started) {
#ifdef __unix
        struct tms t;
        times(&t);
        _start_time = t.tms_utime;
        _started = true;
#else
        _start_time = clock();
#endif
    }
    _running = true;
}


void
Timer::
stop()
{
#ifdef __unix
    struct tms t;
    times(&t);
    _stop_time = t.tms_utime;
#else
    _stop_time = clock();
#endif
    _running = false;
}


double
Timer::
time() const
{
    return static_cast<double>(_stop_time - _start_time)/SC_CLK_TCK;
}


void
Timer::
show(std::ostream& o) const
{
    int diff = _stop_time - _start_time;
    if (_running) {
#ifdef __unix
        struct tms t;
        times(&t);
        diff = t.tms_utime - _start_time;
#else
        diff = clock() - _start_time;
#endif
    }
    int h = diff / 3600 / SC_CLK_TCK;
    int m = (diff / 60 / SC_CLK_TCK) % 60;
    int s = (diff / SC_CLK_TCK) % 60;
    int c = (diff * 100 / SC_CLK_TCK) % 100;

    o << std::setw(2) << std::setfill('0') << h << ":";
    o << std::setw(2) << std::setfill('0') << m << ":";
    o << std::setw(2) << std::setfill('0') << s << ".";
    o << std::setw(2) << std::setfill('0') << c;
}


std::string
Timer::
str() const
{
    std::ostringstream s;
    show(s);
    return s.str();
}


std::ostream&
operator<<(std::ostream& o, const Timer& timer)
{
    timer.show(o);
    return o;
}
}
