/*
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see  <http://www.gnu.org/licenses/>
 */

#include "Problem.hxx"
#include "Reader.hxx"

#include "SFDRFLPAmaralModel.hxx"
#include "DRFLPAmaralModel.hxx"
#include "PROPBetweennessModel.hxx"
#include "DRFLPBetweennessModel.hxx"
#include "DRFLPSAModel.hxx"
#include "SFDRFLPSAModel.hxx"

#include "SFDRFLPBnBSolver.hxx"
#include "DRFLPBnBSolver.hxx"
#include "CplexSolver.hxx"
#include "GurobiSolver.hxx"

#include "BigM.hxx"

#include "CmdArgs.hxx"
#include "Timer.hxx"

#include <iostream>
#include <cstdlib>


namespace {

// show the order of the departments in each row in the solution
void show_layout(std::ostream& out, const std::vector<std::vector<ID>>& rows) {
    for (auto row : rows) {
        out << "[";
        for (size_type i = 0; i < row.size(); i++) {
            if (i > 0) out << ", ";
            out << row[i]+1;
        }
        out << "]\n";
    }
}

}


int main( int argc, const char** argv )
{
    lyutil::CmdArgs args;

    args.add_section("Files:");
    args.add("data", "File containing the data (departments and dependencies).")
        .required( "file-name" )
        .default_value( "a_s8.txt" );

    // problem variant
    args.add_section("Problem type:");
    args.add("drflp", "Solve DRFLP (our approach)");
    args.add("sfdrflp", "Solve SF-DRFLP (our approach)");
    args.add("mrflp", "Solve MRFLP (our approach) with given number of rows").required("NROWS").default_value("3");
    args.add("sfmrflp", "Solve SF-MRFLP (our approach) with given number of rows").required("NROWS").default_value("3");
    args.add("prop", "Solve PROP (two rows) (using our approach)");
    args.add("kprop1", "Solve kPROP, distance type 1");
    args.add("kprop2", "Solve kPROP, distance type 2");
    args.add("kprop3", "Solve kPROP, distance type 3");

    args.add("amaraldrflp", "Solve DRFLP with approach of Amaral");
    args.add("amaralsfdrflp", "Solve SF-DRFLP with approach of Amaral");
    args.add("sadrflp", "Solve DRFLP with approach of Secchin and Amaral");
    args.add("sasfdrflp", "Solve SF-DRFLP with extended approach of Secchin and Amaral");

    args.add_section("Options:");
    args.add("cuts", "Use separation (possible: drflp, sfdrflp, prop, kprop*)");
    args.add("clique", "Use clique constraints (possible: prop)");
    args.add("bigm", "Use improved big-M (possible: drflp, amaraldrflp, sadrflp)");
    args.add("reduce", "Use reduced number of possible assignments (possible: sfdrflp, amaralsfdrflp)");
    args.add("cpx", "Use Cplex MIP solver (default)");
    args.add("grb", "Use Gurobi MIP solver");
    args.add_with_required("tl", "[REAL]",
                           "Time limit in seconds (not with drflp and sfdrflp in enumeration scheme)",
                           "infinity");

    args.add_section("Other options:");
    args.add('h', "help", "Show this help-page.");

    try {
        args.parse(argc, argv);
    }
    catch(lyutil::CmdArgError& e) {
        std::cerr << e.what() << std::endl;
        std::cerr << "Try '" << argv[0] << " --help'" << std::endl;
        std::exit(EXIT_FAILURE);
    }


    if (args.has("help")) {
        std::cout << "Usage: " << argv[0] << " [OPTIONS]" << std::endl
                  << args << std::endl;
        std::exit(EXIT_FAILURE);
    }

    // get problem type
    bool DRFLP = args.has( "drflp" );
    bool SFDRFLP = args.has( "sfdrflp" );
    bool MRFLP = args.has( "mrflp" );
    bool SFMRFLP = args.has( "sfmrflp" );
    bool PROP = args.has( "prop" );
    bool KPROP1 = args.has( "kprop1" );
    bool KPROP2 = args.has( "kprop2" );
    bool KPROP3 = args.has( "kprop3" );
    bool AmaralDRFLP = args.has( "amaraldrflp" );
    bool AmaralSFDRFLP = args.has( "amaralsfdrflp" );
    bool SADRFLP = args.has( "sadrflp" );
    bool SASFDRFLP = args.has( "sasfdrflp" );

    size_type cnt_problems = 0;
    if ( DRFLP ) cnt_problems++;
    if ( SFDRFLP ) cnt_problems++;
    if ( MRFLP ) cnt_problems++;
    if ( SFMRFLP ) cnt_problems++;
    if ( PROP ) cnt_problems++;
    if ( KPROP1 ) cnt_problems++;
    if ( KPROP2 ) cnt_problems++;
    if ( KPROP3 ) cnt_problems++;
    if ( AmaralDRFLP ) cnt_problems++;
    if ( AmaralSFDRFLP ) cnt_problems++;
    if ( SADRFLP ) cnt_problems++;
    if ( SASFDRFLP ) cnt_problems++;

    if ( cnt_problems == 0 ){
        std::cerr << "No problem type selected!" << std::endl;
        std::exit(EXIT_FAILURE);
    }
    if ( cnt_problems > 1 ){
        std::cerr << "More than one problem type selected!" << std::endl;
        std::exit(EXIT_FAILURE);
    }

    // get options
    bool use_cuts = args.has("cuts");
    if (use_cuts && !(DRFLP || SFDRFLP || MRFLP || SFMRFLP || PROP || KPROP1 || KPROP2 || KPROP3 )) {
        std::cerr << "Option 'cuts' not allowed with --amaral* and --sa models!" << std::endl;
        std::exit(EXIT_FAILURE);
    }

    bool use_cliques = args.has("clique");
    if (use_cliques && !(PROP || KPROP1 || KPROP2 || KPROP3) ) {
        std::cerr << "Option 'clique' requires --prop or --kprop{1,2,3}!" << std::endl;
        std::exit(EXIT_FAILURE);
    }

    bool use_bigm = args.has("bigm");
    if (use_bigm && !(DRFLP || MRFLP || AmaralDRFLP || SADRFLP ) ){
        std::cerr << "Option 'bigm' requires --drflp or --mrflp or --amaraldrflp or --sadrflp!" << std::endl;
        std::exit(EXIT_FAILURE);
    }

    bool use_reduce = args.has("reduce");
    if (use_reduce && !(SFDRFLP || SFMRFLP) ) {
        std::cerr << "Option 'reduce' does not fit to problem type (only --sfdrflp and --sfmrflp)!" << std::endl;
        std::exit(EXIT_FAILURE);
    }

    bool use_grb = args.has("grb");
    bool use_cpx = args.has("cpx") || !use_grb;
    if (use_cpx && use_grb) {
        std::cerr << "Only one MIP solver can be used." << std::endl;
        std::exit(EXIT_FAILURE);
    }

    double time_limit = (args.has("tl")
                         ? std::atof(args.get("tl").c_str())
                         : std::numeric_limits<double>::infinity());
    if (time_limit < std::numeric_limits<double>::infinity() && (SFDRFLP || DRFLP || SFMRFLP || MRFLP) ) {
        std::cerr << "Option 'tl' is not possible with DRFLP, MRFLP and SFDRFLP, SFMRFLP (enumeration scheme)!" << std::endl;
        std::exit(EXIT_FAILURE);
    }

    // whether the instance file must contain row information
    bool has_rows = !(DRFLP || SFDRFLP || MRFLP || SFMRFLP || AmaralDRFLP || AmaralSFDRFLP || SADRFLP || SASFDRFLP );

    // set input file
    Reader reader(args.get("data"), has_rows);

    // read data
    std::unique_ptr<Problem> problem = reader.read_problem();

    // compute big-M
    double bigm = (use_bigm
                   ? compute_improved_bigm(*problem)
                   : compute_simple_bigm(*problem));

    lyutil::Timer timer(true);
    double objval = 0;

    std::unique_ptr<MIPSolver> solver;
    if (use_cpx)
        solver.reset(new CplexSolver(time_limit));
    else
        solver.reset(new GurobiSolver(time_limit));

    // solve row layout problem depending on the type
    if (DRFLP) {
        DRFLPBnBSolver bnbsolver(*problem, use_cuts, bigm, *solver);
        objval = bnbsolver.solve();
        timer.stop();
        std::cout << "Objective: " << objval << std::endl;
        std::cout << "Time: " << timer << std::endl;
        show_layout(std::cout, bnbsolver.get_solution());
    }
    else if (SFDRFLP) {
        SFDRFLPBnBSolver bnbsolver(*problem, use_cuts, use_reduce, *solver);
        objval = bnbsolver.solve();
        timer.stop();
        std::cout << "Objective: " << objval << std::endl;
        std::cout << "Time: " << timer << std::endl;
        show_layout(std::cout, bnbsolver.get_solution());
    }
    else if (MRFLP) {
        DRFLPBnBSolver bnbsolver(*problem, std::stoul(args.get("mrflp")), use_cuts, bigm, *solver);
        objval = bnbsolver.solve();
        timer.stop();
        std::cout << "Objective: " << objval << std::endl;
        std::cout << "Time: " << timer << std::endl;
        show_layout(std::cout, bnbsolver.get_solution());
    }
    else if (SFMRFLP) {
        SFDRFLPBnBSolver bnbsolver(*problem, std::stoul(args.get("sfmrflp")), use_cuts, use_reduce, *solver);
        objval = bnbsolver.solve();
        timer.stop();
        std::cout << "Objective: " << objval << std::endl;
        std::cout << "Time: " << timer << std::endl;
        show_layout(std::cout, bnbsolver.get_solution());
    }
    else if (PROP) {
        PROPBetweennessModel model(*problem,
                                   problem->get_rows(),
                                   use_cuts,
                                   use_cliques,
                                   Cost1);
        model.init();
        objval = solver->solve(model);
        std::cout << "Lower bound: " << solver->get_lower_bound() << std::endl;
        timer.stop();
        std::cout << "Objective: " << objval << std::endl;
        std::cout << "Time: " << timer << std::endl;
        show_layout(std::cout, model.compute_solution( solver->get_solution() ));
    }
    else if (KPROP1) {
        PROPBetweennessModel model(*problem,
                                   problem->get_rows(),
                                   use_cuts,
                                   use_cliques,
                                   Cost1);
        model.init();
        objval = solver->solve(model);
        std::cout << "Lower bound: " << solver->get_lower_bound() << std::endl;
        timer.stop();
        std::cout << "Objective: " << objval << std::endl;
        std::cout << "Time: " << timer << std::endl;
        show_layout(std::cout, model.compute_solution( solver->get_solution() ));
    }
    else if (KPROP2) {
        PROPBetweennessModel model(*problem,
                                   problem->get_rows(),
                                   use_cuts,
                                   use_cliques,
                                   Cost2);
        model.init();
        objval = solver->solve(model);
        objval += model.compute_fixed_costs();
        std::cout << "Lower bound: " << solver->get_lower_bound()+model.compute_fixed_costs() << std::endl;
        timer.stop();
        std::cout << "Objective: " << objval << std::endl;
        std::cout << "Time: " << timer << std::endl;
        show_layout(std::cout, model.compute_solution( solver->get_solution() ));
    }
    else if (KPROP3) {
        PROPBetweennessModel model(*problem,
                                   problem->get_rows(),
                                   use_cuts,
                                   use_cliques,
                                   Cost3);
        model.init();
        objval = solver->solve(model);
        std::cout << "Lower bound: " << solver->get_lower_bound() << std::endl;
        timer.stop();
        std::cout << "Objective: " << objval << std::endl;
        std::cout << "Time: " << timer << std::endl;
        show_layout(std::cout, model.compute_solution( solver->get_solution() ));
    }
    else if (AmaralDRFLP) {
        DRFLPAmaralModel model(*problem, use_cuts, bigm);
        model.init();
        objval = solver->solve(model);
        std::cout << "Lower bound: " << solver->get_lower_bound() << std::endl;
        timer.stop();
        std::cout << "Objective: " << objval << std::endl;
        std::cout << "Time: " << timer << std::endl;
        show_layout(std::cout, model.compute_solution( solver->get_solution() ));
    }
    else if (AmaralSFDRFLP) {
        SFDRFLPAmaralModel model(*problem, use_cuts);
        model.init();
        objval = solver->solve(model);
        std::cout << "Lower bound: " << solver->get_lower_bound() << std::endl;
        timer.stop();
        std::cout << "Objective: " << objval << std::endl;
        std::cout << "Time: " << timer << std::endl;
        show_layout(std::cout, model.compute_solution( solver->get_solution() ));
    }
    else if (SADRFLP) {
        DRFLPSAModel model(*problem, use_cuts, bigm);
        model.init();
        objval = solver->solve(model);
        std::cout << "Lower bound: " << solver->get_lower_bound() << std::endl;
        timer.stop();
        std::cout << "Objective: " << objval << std::endl;
        std::cout << "Time: " << timer << std::endl;
        show_layout(std::cout, model.compute_solution( solver->get_solution() ));
    }
    else if (SASFDRFLP) {
        SFDRFLPSAModel model(*problem, use_cuts);
        model.init();
        objval = solver->solve(model);
        std::cout << "Lower bound: " << solver->get_lower_bound() << std::endl;
        timer.stop();
        std::cout << "Objective: " << objval << std::endl;
        std::cout << "Time: " << timer << std::endl;
        show_layout(std::cout, model.compute_solution( solver->get_solution() ));
    }

    return 0;
}
